/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include "keyer_pins.h"
#include "keyer_beep.h"
#include "keyer_config.h"
#include "keyer_control.h"
#include "keyer_encoder.h"
#include "keyer_keyer.h"
#include "keyer_messages.h"
#include "keyer_mpr121.h"
#include "keyer_outputs.h"
#include "keyer_paddle.h"
#include "keyer_protocol.h"
#include "keyer_queue.h"
#include "keyer_status.h"
#include "keyer_comm.h"

// This is the ten dollar keyer.  It is a low-cost keyer with a built-in
// paddle and sidetone.  It is designed for code practice but it also
// can key a radio and work with most PC logging programs.
//
// The keyer software is built around a simple message queue.  Much
// of the communication between classes and all communication from
// interrupt handlers is done by queuing messages.  This makes it
// easy to separate the hardware dependent code from the independent
// code in most cases.
//
// All inputs are interrupt-driven.  No hardware is polled in the main
// loop.
//
// This is written as an Arduino program.  The various classes are
// set up in the setup() routine.  Then loop() is called in a loop.
// The Arduino environment supports C++ but the runtime library is
// limited.  And dynamic memory (malloc() and free() or new and
// delete) are not allowed.  These are reasonable restrictions given
// the memory size - as it is the libraries take up a lot of the
// code space.
//
// Support for the STM32F103C8T6 processor in the Arduino environment
// is available from www.stm32duino.com
//
// Coding conventions are simple:  Indents are four spaces (same as
// the Maple libraries it uses).  Variables use underscores rather
// than camel-case.  Class variables start with m_.  That's about it.
// The ?: conditional syntax isn't used - too many people mess it up.
//
// *******************************************************************
// **                                                               **
// **   The edit history and version number are in keyer_version.h  **
// **   If you modify the code update that file.                    **
// **                                                               **
// *******************************************************************

#define BAUD 1200

void
setup()
//----------------------------------------------------------------------
// Set up the various components
//----------------------------------------------------------------------
{
    // initialize the on-board LED as an output.  For now it follows
    // PTT but it can be used for debugging
    pinMode(LED, OUTPUT);
    digitalWrite(LED, HIGH);

    Serial.begin(BAUD);  // BAUD has no effect on USB serial: placeholder for physical UAR
    
    comm.setup();

    config.setup();

#ifdef DEBUG
    config.display();
#endif

    // There are some dependencies in the setup routines.  Don't change
    // the order without checking.
    outputs.setup();
    beep.setup();
    protocol.setup();
    keyer.setup();
    status.setup();
    control.setup();
    mpr121.setup();
    paddle.setup();

    // There is a bug in the timer code - it needs to fire once.  Wait before
    // finishing setup.
    delay(1000);

    // From here on the Arduino timer is not used.  It could be turned off.
    
    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "OK");
}

void
loop()
//----------------------------------------------------------------------
// Main program loop - service the queue
//----------------------------------------------------------------------
{
    // Items are placed on the queue by interrupt routines.  It is
    // probably possible to sleep between queue runs as the interrupts
    // will wake the processor. 
    queue.run();
}


