/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include "keyer_pins.h"
#include "keyer_config.h"
#include "keyer_beep.h"
#include "keyer_error.h"
#include "keyer_messages.h"
#include "keyer_misc.h"
#include "keyer_queue.h"

// The tone generator is used by the keyer and by the status keyer.  It
// can only output one tone at a time.  A member variable controls which
// keyer "owns" the sidetone.

// The STM32F103C8T6 does not have a digital to analog converter.  Generate
// a somewhat-sine wave by using pulse width modulation.  The keyer has a
// simple low pass filter which will smooth the waveform somewhat.  This
// DSP method is inexpensive and sounds much better than a square wave - and
// the volume can be controlled.

// Timer 2 and DMA are used to produce a sineusoidal sidetone.  The CPU is
// not involved except for set-up.  The speaker and  headphones are on
// different pins with different timer channels and therefore different DMA
// channels.  The pinMode routine doesn't set up the timer in the way this
// code needs so it uses the gpio routine and sets up the timer and DMA
// separately. 

// This class is harware dependent.  It depends on the PWM and DMA
// capabilities of the STM32F103 series microprocessors.

// Sine wave tables as a ratio.  This is half of a cycle.  The tables are
// for 32 and 64 samples per cycle.
static const uint32_t sine_wave_64[32] = {
    0x00000000, 0x0c8bd35e, 0x18f8b83c, 0x25280c5d, 0x30fbc54d,
    0x3c56ba70, 0x471cece6, 0x5133cc94, 0x5a827999, 0x62f201ac,
    0x6a6d98a3, 0x70e2cbc5, 0x7641af3c, 0x7a7d055a, 0x7d8a5f3f,
    0x7f62368e, 0x7fffffff, 0x7f62368e, 0x7d8a5f3f, 0x7a7d055a,
    0x7641af3c, 0x70e2cbc5, 0x6a6d98a3, 0x62f201ac, 0x5a827999,
    0x5133cc94, 0x471cece6, 0x3c56ba70, 0x30fbc54d, 0x25280c5d,
    0x18f8b83c, 0x0c8bd35e
};

static const uint32_t sine_wave_32[16] = {
    0x00000000, 0x18f8b83c, 0x30fbc54d, 0x471cece6, 0x5a827999,
    0x6a6d98a3, 0x7641af3c, 0x7d8a5f3f, 0x7fffffff, 0x7d8a5f3f,
    0x7641af3c, 0x6a6d98a3, 0x5a827999, 0x471cece6, 0x30fbc54d,
    0x18f8b83c
};

// Table of dB as a ratio.  All numbers represent -1 dB steps.
static const uint32_t dB[61] = {
    0xffffffff, 0xe429057f, 0xcb59185e, 0xb53bef57, 0xa1866ba7,
    0x8ff599e0, 0x804dce79, 0x7259db1c, 0x65ea59fd, 0x5ad50cdf,
    0x50f44d89, 0x48268deb, 0x404de61f, 0x394faecf, 0x331426af,
    0x2d8621c7, 0x2892c18b, 0x242934c0, 0x203a7e5b, 0x1cb94299,
    0x1999999a, 0x16d0e6f3, 0x1455b5a3, 0x121f97ef, 0x10270ac4,
    0x0e655c30, 0x0cd494a6, 0x0b6f62b6, 0x0a310900, 0x09154e16,
    0x08186e27, 0x07370e31, 0x066e309d, 0x05bb2b15, 0x051b9d78,
    0x048d69c7, 0x040eacf4, 0x039db87a, 0x03390ca3, 0x02df5376,
    0x028f5c29, 0x02481718, 0x0208922a, 0x01cff598, 0x019d8114,
    0x01708938, 0x01487544, 0x0124bd12, 0x0104e74d, 0x00e887cf,
    0x00cf3e37, 0x00b8b49f, 0x00a49e76, 0x0092b782, 0x0082c2f2,
    0x00748a94, 0x0067de18, 0x005c9273, 0x00528143, 0x00498859,
    0x00418937
};

// The volume is in dB, where larger numbers are quieter.
// Zero is the maximum level.
static const int DB_MAX = 60;
static const int DB = DB_MAX - 1;

// The sidetone frequencies are limited.  If it is desired to
// go lower a 128 sample table should be added or two cycles
// per sample should be implemented in order to keep the switching
// frequency high.  Higer frequencies could be output but the
// low pass filter will start to attenuate them.  And nobody needs
// a real high pitched sidetone.
static const int MIN_FREQ = 300;
static const int MAX_FREQ = 1200;
static const int INC_FREQ = 10;

// The amplifier is shutdown when the output is high
static const uint8_t AMPLIFIER_ON = LOW;
static const uint8_t AMPLIFIER_OFF = HIGH;

keyer_beep::keyer_beep()
{
    // GPIOA Periph clock enable
    rcc_clk_enable(RCC_GPIOA);

    // Set the amplifier pin as an output and turned off.
    pinMode(AMPLIFIER_SHUTDOWN, OUTPUT_OPEN_DRAIN);
    digitalWrite(AMPLIFIER_SHUTDOWN, AMPLIFIER_OFF);

    // These values are "impossible".  They will be set by the first
    // set_output call.
    m_freq = 0xffffffff;
    m_ampl = 0xffffffff;
    m_phones = false;

    m_owner = OWNER_KEYER;
    m_beeping = false;
}

void
keyer_beep::setup()
//----------------------------------------------------------------------
// Set up the tone generator
//----------------------------------------------------------------------
{
    queue.subscribe(CLASS_BEEP, this);
    set_output();
}

void
keyer_beep::set_output(uint32_t freq, uint32_t ampl, bool phones)
//----------------------------------------------------------------------
// Set up the tone generator to the specified freq, volume, & output
//----------------------------------------------------------------------
{
    uint8_t channel;
    timer_dma_base_addr dma_base;

    uint32_t ratio;
    const uint32_t *sine_wave;

    uint32_t sample_ticks;
    uint32_t ticks;
        
    // If nothing has changed just return
    if ((freq == m_freq) && (ampl == m_ampl) && (phones == m_phones)) {
        return;
    }

    if (ampl) {

        digitalWrite(AMPLIFIER_SHUTDOWN, AMPLIFIER_ON);

        // This doesn't use pinMode() because that routine changes the timer.
        if (phones) {
            gpio_set_mode(PIN_MAP[HEADPHONE_AUDIO].gpio_device,
                          PIN_MAP[HEADPHONE_AUDIO].gpio_bit, GPIO_AF_OUTPUT_PP);
            gpio_set_mode(PIN_MAP[SPEAKER_AUDIO].gpio_device,
                          PIN_MAP[SPEAKER_AUDIO].gpio_bit, GPIO_OUTPUT_PP);

            channel = TIMER_CH1;
            m_dchannel = DMA_CH5;
            dma_base = TIMER_DMA_BASE_CCR1;
        }
        else {
            gpio_set_mode(PIN_MAP[HEADPHONE_AUDIO].gpio_device,
                          PIN_MAP[HEADPHONE_AUDIO].gpio_bit, GPIO_OUTPUT_PP);
            gpio_set_mode(PIN_MAP[SPEAKER_AUDIO].gpio_device,
                          PIN_MAP[SPEAKER_AUDIO].gpio_bit, GPIO_AF_OUTPUT_PP);

            channel = TIMER_CH2;
            m_dchannel = DMA_CH7;
            dma_base = TIMER_DMA_BASE_CCR2;
        }

        // Get the ratio for the amplitude
        if (ampl > DB_MAX) {
            ampl = DB_MAX;
        }

        // The db table is writen maximum to minimum, and an amplitude of zero is off.
        ratio = dB[DB_MAX - ampl];
    
        // The sampling frequency must be above the audio range.  But
        // larger numbers of samples at high frequencies cut down the
        // number of bits of resolution.  Use 32 samples per cycle above
        // 600 Hz and 64 samples below.  Don't go below 300 Hz.
        if (freq < MIN_FREQ) {
            freq = MIN_FREQ;
        }
        if (freq > MAX_FREQ) {
            freq = MAX_FREQ;
        }

        if (freq < 600) {
            m_samples = 64;
            sine_wave = sine_wave_64;
        }
        else {
            m_samples = 32;
            sine_wave = sine_wave_32;
        }

        // Canculate the number of ticks for the ARR.  The number is rounded by adding freq/2.
        sample_ticks = CLOCK_HZ / m_samples;
        ticks = ((sample_ticks + (freq / 2)) / freq);

        // Calculate the CCR values for the sine wave samples
        for (int i=0; i<m_samples/2; i++) {
            uint32_t v = (((long long)(sine_wave[i]) * ratio) >> 32);
            uint32_t v1 = 0x80000000 + v;
            uint32_t v2 = 0x80000000 - v;
            m_ccr[i] = ((unsigned long long)ticks * v1) >> 32;
            m_ccr[i+(m_samples/2)] = ((unsigned long long)ticks * v2) >> 32;
        }

        dma_init(DMA1);
        keyer_dma_setup_transfer(DMA1, m_dchannel, (volatile void*)&(timer2.regs.gen->DMAR), DMA_SIZE_16BITS, 
                                (volatile void*)m_ccr, DMA_SIZE_16BITS,
                                DMA_FROM_MEM | DMA_MINC_MODE | DMA_CIRC_MODE);
        dma_set_priority(DMA1, m_dchannel, DMA_PRIORITY_HIGH);
        dma_set_num_transfers(DMA1, m_dchannel, m_samples);
            
        timer_set_prescaler(TIMER2, 0);
        timer_set_mode(TIMER2, channel, TIMER_PWM);
        timer_set_reload(TIMER2, ticks);
        timer_set_compare(TIMER2, channel, ticks / 2);

        timer_oc_set_mode(TIMER2, channel, TIMER_OC_MODE_PWM_1, TIMER_OC_PE);
        dma_clear_isr_bits(DMA1, m_dchannel);

        timer_dma_set_base_addr(TIMER2, dma_base);
        timer_dma_enable_req(TIMER2, channel);
        timer_resume(TIMER2);

        if (m_beeping) {
            dma_enable(DMA1, m_dchannel);
        }
    }
    else {
        // Turning the amplifier on takes a significant fraction of a
        // second.  So only turn it off if both keyers have no beep.
        bool amplifier = false;

        if (config.get_use_phones()) {
            if (config.get_headphone_vol()) {
                amplifier = true;
            }
        }
        else {
            if (config.get_speaker_vol()) {
                amplifier = true;
            }
        }

        if (config.get_xuse_phones()) {
            if (config.get_xheadphone_vol()) {
                amplifier = true;
            }
        }
        else {
            if (config.get_xspeaker_vol()) {
                amplifier = true;
            }
        }

        if (!amplifier) {
            digitalWrite(AMPLIFIER_SHUTDOWN, AMPLIFIER_OFF);
        }
        keyer_dma_uninit(DMA1);
        m_beeping = false;
    }

    m_freq = freq;
    m_ampl = ampl;
    m_phones = phones;
}

void
keyer_beep::set_output()
//----------------------------------------------------------------------
// Set up the tone generator to the configured values
//----------------------------------------------------------------------
{
    if (m_owner == OWNER_KEYER) {
        uint8_t use_phones = config.get_use_phones();
        uint8_t ampl;
        if (use_phones) {
            ampl = config.get_headphone_vol();
        }
        else {
            ampl = config.get_speaker_vol();
        }
        set_output(config.get_frequency()*10, ampl, use_phones);
    }
    else { // OWNER_STATUS
        uint8_t use_phones = config.get_xuse_phones();
        uint8_t ampl;
        if (use_phones) {
            ampl = config.get_xheadphone_vol();
        }
        else {
            ampl = config.get_xspeaker_vol();
        }
        
        set_output(config.get_xfrequency()*10, ampl, use_phones);
    }
}

void
keyer_beep::on()
//----------------------------------------------------------------------
// Turn on the tone generator
//----------------------------------------------------------------------
{
    dma_disable(DMA1, m_dchannel);
    keyer_dma_set_circular(DMA1, m_dchannel);
    dma_set_num_transfers(DMA1, m_dchannel, m_samples);
    dma_enable(DMA1, m_dchannel);

    m_beeping = true;
}

void
keyer_beep::off()
//----------------------------------------------------------------------
// Turn off the tone generator
//----------------------------------------------------------------------
{
   keyer_dma_clear_circular(DMA1, m_dchannel);

   m_beeping = false;
}


void
keyer_beep::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle tone generator messages
//----------------------------------------------------------------------
{  
    switch (MESSAGE_ID(message)) {
        case MSG_BEEP_ON:
        {
            if (data != m_owner) {
                return;
            }

            if (!m_ampl) {
                return;
            }

            on();
            break;
        }
        
        case MSG_BEEP_OFF:
        {
            if (data != m_owner) {
                return;
            }

            if (!m_ampl) {
                return;
            }

            off();
            break;
        }

       case MSG_BEEP_OWNER:
        {
            if (data == m_owner) {
                return;
            }

            m_owner = data;

            if (m_beeping) {
                off();
            }

            set_output();
            break;
        }

        case MSG_BEEP_UPDATE:
        {
            if (data == m_owner) {
                set_output();
            }
            break;
        }

       case MSG_FREQUENCY_UP:
        {
            if (data == OWNER_KEYER) {
                uint8_t freq = config.get_frequency();
                if (freq < (MAX_FREQ / 10)) {
                    freq++;
                    config.set_frequency(freq);
                    if (m_owner == OWNER_KEYER) {
                        set_output();
                    }
                }
            }
            else { // OWNER_STATUS
                uint8_t freq = config.get_xfrequency();
                if (freq < (MAX_FREQ / 10)) {
                    freq++;
                    config.set_xfrequency(freq);
                    if (m_owner == OWNER_STATUS) {
                        set_output();
                    }
                }
            }

            break;
        }

       case MSG_FREQUENCY_DOWN:
        {
            if (data == OWNER_KEYER) {
                uint8_t freq = config.get_frequency();
                if (freq > (MIN_FREQ / 10)) {
                    freq--;
                    config.set_frequency(freq);
                    if (m_owner == OWNER_KEYER) {
                        set_output();
                    }
                }
            }
            else { // OWNER_STATUS
                uint8_t freq = config.get_xfrequency();
                if (freq > (MIN_FREQ / 10)) {
                    freq--;
                    config.set_xfrequency(freq);
                    if (m_owner == OWNER_STATUS) {
                        set_output();
                    }
                }
            }

            break;
        }

       case MSG_VOLUME_UP:
        {
            uint8_t use_phones = config.get_use_phones();
            if (data == OWNER_KEYER) {
                uint8_t ampl;
                if (use_phones) {
                    ampl = config.get_headphone_vol();
                }
                else {
                    ampl = config.get_speaker_vol();
                }
                if (ampl < DB_MAX) {
                    ampl++;
                    if (use_phones) {
                        config.set_headphone_vol(ampl);
                    }
                    else {
                        config.set_speaker_vol(ampl);
                    }
                    if (m_owner == OWNER_KEYER) {
                        set_output();
                    }
                }
            }
            else { // OWNER_STATUS
                uint8_t ampl;
                if (use_phones) {
                    ampl = config.get_xheadphone_vol();
                }
                else {
                    ampl = config.get_xspeaker_vol();
                }
                if (ampl < DB_MAX) {
                    ampl++;
                    if (use_phones) {
                        config.set_xheadphone_vol(ampl);
                    }
                    else {
                        config.set_xspeaker_vol(ampl);
                    }
                    if (m_owner == OWNER_STATUS) {
                        set_output();
                    }
                }
            }

            break;
        }

       case MSG_VOLUME_DOWN:
        {
            uint8_t use_phones = config.get_use_phones();
            if (data == OWNER_KEYER) {
                uint8_t ampl;
                if (use_phones) {
                    ampl = config.get_headphone_vol();
                }
                else {
                    ampl = config.get_speaker_vol();
                }
                if (ampl > 0) {
                    ampl--;
                    if (use_phones) {
                        config.set_headphone_vol(ampl);
                    }
                    else {
                        config.set_speaker_vol(ampl);
                    }
                    if (m_owner == OWNER_KEYER) {
                        set_output();
                    }
                }
            }
            else { // OWNER_STATUS
                uint8_t ampl;
                if (use_phones) {
                    ampl = config.get_xheadphone_vol();
                }
                else {
                    ampl = config.get_xspeaker_vol();
                }
                if (ampl > 0) {
                    ampl--;
                    if (use_phones) {
                        config.set_xheadphone_vol(ampl);
                    }
                    else {
                        config.set_xspeaker_vol(ampl);
                    }
                    if (m_owner == OWNER_STATUS) {
                        set_output();
                    }
                }
            }

            break;
        }

        case MSG_VALUES:
        {
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_CLEAR), 0);
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_NUMBER), config.get_frequency()*10);
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_CHAR), '/');
            if (config.get_use_phones()) {
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_NUMBER), config.get_headphone_vol());
            }
            else {
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_NUMBER), config.get_speaker_vol());
            }

            break;
        }

        case MSG_VALUES_XVOL:
        {
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_CLEAR), 0);
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_NUMBER),
                      config.get_xspeaker_vol());

            break;
        }

        case MSG_VALUES_XFREQ:
        {
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_CLEAR), 0);
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_NUMBER),
                      config.get_xfrequency()*10);

            break;
        }

        case MSG_RECALC:
        {
            set_output();
            break;
        }

        default:
            error.bad_message("Beep", message, data);
    }
}

keyer_beep beep;

