/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include <wirish.h>

#include "keyer_error.h"
#include "keyer_keyer.h"
#include "keyer_queue.h"
#include "keyer_bug.h"

// A "bug" is a mechanical semi-automatic keyer.  It generates dits
// automatically and dahs are manual.  The automatic dit mechanism
// does not produce self-completing dits - if the paddle is released
// while a dit is being sent that dit will be truncated.  Also the
// power for the automatic mechanism is supplied by the force of the
// user pressing the paddle.  That force is finite and eventually the
// automatic mechanism will run out of power.  Both of these aspects are
// emulated in this code.  With a single-lever paddle this should behave
// reasonably similarly to a physical bug.

// This provides the manual (paddle) control for the keyer in bug mode.
// The code in keyer_keyer calls this through the public routines
// inherited from keyer_manual.  And this code calls the keyer through
// the public routines in that code.

// This class is hardware independent.

keyer_bug::keyer_bug()
{
    
}

bool
keyer_bug::setup(keyer_type type, bool silent)
//----------------------------------------------------------------------
// If the keyer needs a "bug" set up to handle it
//----------------------------------------------------------------------
{
    if (type != KEYER_TYPE_BUG) {
        return false;
    }

    if (!silent) {
        queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "BUG");
    }

    m_cap_left = false;
    m_cap_right = false;
    m_phy_left = false;
    m_phy_right = false;

    m_dit_on = false;
    m_dah_on = false;

    m_sending_dit = false;
    m_sending_dah = false;

    m_dits = 0;

    return true;
}

void
keyer_bug::paddle_change(uint32_t id, uint32_t data)
//----------------------------------------------------------------------
// Handle a paddle opening or closing
//----------------------------------------------------------------------
{
    // Any paddle change wipes out the dit decay
    keyer.km_alter_weight(0);
    
    switch (id) {
        case MSG_CAP_LEFT_CHANGED:
        {
            if (data) {
                m_cap_left = true;
            }
            else {
                m_cap_left = false;
            }
    
            break;
        }

        case MSG_CAP_RIGHT_CHANGED:
        {
            if (data) {
                m_cap_right = true;
            }
            else {
                m_cap_right = false;
            }
    
            break;
        }

        case MSG_PHY_LEFT_CHANGED:
        {
            if (data) {
                m_phy_left = true;
            }
            else {
                m_phy_left = false;
            }
    
            break;
        }
 
        case MSG_PHY_RIGHT_CHANGED:
        {
            if (data) {
                m_phy_right = true;
            }
            else {
                m_phy_right = false;
            }
    
            break;
        }
    }

    if (config.get_keyer_paddle_swap()) {
        m_dit_on = m_cap_right || m_phy_right;
        m_dah_on = m_cap_left || m_phy_left;
    }
    else {
        m_dit_on = m_cap_left || m_phy_left;
        m_dah_on = m_cap_right || m_phy_right;
    }

    // A bug is a single-lever device.  The handling of both paddles being
    // pressed doesn't match any actual bug.  Arbitrarily the dah has priority.
    if (m_dit_on || m_dah_on) {
        if (m_dah_on) {
            m_sending_dit = false; 
            if (!m_sending_dah) {
                keyer.km_start_sending();
                m_sending_dah = true;
            }
        }
        else {
            if (m_sending_dah) {
                keyer.km_stop_sending();
                m_sending_dah = false;
            }

            if (!m_sending_dit) {
                keyer.km_send_dit();
                m_sending_dit = true;
                m_dits = 0;
            }
        }
    }
    else {
        // Real bugs do not have self-completing dits.  They stop
        // immediately when the paddle is released.
        keyer.km_stop_sending();
        m_sending_dit = false;
        m_sending_dah = false;
    }
}

void
keyer_bug::state_change(keyer_state state)
//----------------------------------------------------------------------
// Handle a keyer state change
//----------------------------------------------------------------------
{

    switch (state) {

        case KEYER_STATE_IDLE:
        {
            if (m_dah_on) {
                keyer.km_start_sending();
                m_sending_dah = true;
                m_sending_dit = false;
            }
            else {
                if (m_sending_dah) {
                    m_sending_dah = false;
                }

                // Somewhat hoakey decay for a mechanical bug.   An attempt to meaasure
                // the actual decay of a bug failed because the sidetone recording
                // had too much jitter due to the sidetone generator.  It should
                // be re-measured.
                if (m_dit_on) {
                    m_dits++;
                    if (m_dits == 11) {
                        keyer.km_alter_weight(10);
                    }
                    if (m_dits == 12) {
                        keyer.km_alter_weight(15);
                    }
                    if (m_dits == 13) {
                        keyer.km_alter_weight(20);
                    }
                    if (m_dits == 14) {
                        keyer.km_alter_weight(25);
                    }
                    if (m_dits == 15) {
                        keyer.km_alter_weight(30);
                    }
                    if (m_dits == 16) {
                        keyer.km_alter_weight(35);
                    }
                    if (m_dits == 17) {
                        keyer.km_alter_weight(40);
                    }

                    if (m_dits < 18) {
                        keyer.km_send_dit();
                    }
                    else {
                        keyer.km_start_sending();
                    }

                    m_sending_dit = true;
                }
            }
            break;
        }

        default:
            break;
    }
}

keyer_bug bug;
