/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include <libmaple/usb_cdcacm.h>

#include "keyer_comm.h"
#include "keyer_error.h"
#include "keyer_messages.h"
#include "keyer_misc.h"
#include "keyer_queue.h"
#include "keyer_usb.h"

// This handles the USB communications.  It gets an interupt when
// characters are available from the PC.  It reads the characters
// and queues messages to the serial protocol handler.  It also
// sends messages from the serial protocol to the PC.
//
// The Arduino / Maple library Serial implementation uses the same
// underlying USB code so it is safe to use the Serial.print()
// family of calls to print debugging information.

// This class is hardware dependent.  It depends on the USB
// library code which is dependent on the USB hardware.

keyer_comm::keyer_comm()
{
  
}

void
keyer_comm::setup()
//----------------------------------------------------------------------
// Set up the USB serial protocol interrupt hooks and callbacks
//----------------------------------------------------------------------
{
    queue.subscribe(CLASS_COMM, this);
    usb_cdcacm_set_hooks(KEYER_USB_CDCACM_HOOK_RX, data_available);    
}

void keyer_comm::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle USB communications messages
//----------------------------------------------------------------------
{
    switch (MESSAGE_ID(message)) {
    
        case MSG_DATA_AVAILABLE:
        {
            while (usb_cdcacm_data_available()) {
                // Don't flood the queue with USB data.
                if (queue.almost_full()) {
                    // There are probably characters from the PC already in the
                    // queue.  When they have been processed come here again.
                    queue.put(MESSAGE(CLASS_COMM, MSG_DATA_AVAILABLE), 0);
                    break;
                }
                uint8_t b;
                if (usb_cdcacm_rx(&b, 1)) {
                    queue.put(MESSAGE(CLASS_PROTOCOL, MSG_COMM_CHAR), b);
                }
                else {
                    // This shouldn't happen.  The buffer size said there was a
                    // character.
                    break;
                }
            }
            break;
        }

        case MSG_SEND_CHAR:
        {
            if (usb_is_configured(USBLIB)) {
                uint8_t b = data;
                usb_cdcacm_tx(&b, 1);
            }
            break;
        }

        case MSG_SEND_NAME:
        {
            if (usb_is_configured(USBLIB)) {
                uint8_t buffer[4];
                buffer[0] = data >> 24;
                buffer[1] = (data >> 16) & 0xFF;
                buffer[2] = (data >> 8) & 0xFF;
                buffer[3] = data & 0xFF;
    
                usb_cdcacm_tx(buffer, 4);
            }
            break;
        }

        default:
            error.bad_message("Comm", message, data);
    }
}

void
keyer_comm::data_available(unsigned int ignore1, void *ignore2)
//----------------------------------------------------------------------
// USB Interrupt callback for when data is available to be read.
//----------------------------------------------------------------------
{
    UNUSED(ignore1);
    UNUSED(ignore2);

    queue.put(MESSAGE(CLASS_COMM, MSG_DATA_AVAILABLE), 0);
}

keyer_comm comm;
