/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include <string.h>

#include "keyer_common.h"
#include "keyer_config.h"
#include "keyer_error.h"
#include "keyer_messages.h"
#include "keyer_queue.h"

#include "flash_stm32.h"

// Various configuration parameters are stored in a table.  They can be accessed
// and changed through the PC interface.  Some can also be changed through using
// the encoders on the keyer.
//
// All configuration values are one byte.
//
// A table in keyer_config.h is processed through various macros to generate
// storage, getters and setters, names etc.

// The config is stored in the last page of flash memory.  Since it is only a few
// bytes long and the page is 1 K each time it is written it is stored higher in
// memory.  Eventually it won't fit - then the page is erased and it goes back to
// the beginning of the page.  This method reduces the number of erase and write
// cycles on the flash.

// Flash is accessed in 16 bit chunks.  If the configuration data is an odd number
// of bytes one extra byte will be written to flash.

// This class is hardware independent except for the code which reads and writes
// the flash memory which is dependent on the hardware flash writer.  The flash
// code should probably be moved into a separate class.

static const uint32_t FLASH_PAGE = 0x0800FC00;
static const uint32_t FLASH_PAGE_SIZE = 1024;
static const uint32_t FLASH_PAGE_END = (FLASH_PAGE + FLASH_PAGE_SIZE - 1);
static const uint32_t FLASH_DATA_SIZE = (sizeof(config_data) + 1) & 0xfffffffe;

enum config_types {
    T_NONE,
    T_SPEED,
    T_FREQ,
    T_VOL,
    T_WEIGHT,
    T_0255,
    T_ONOFF
};

#define CV(v_name, def_val, ext_name, string_name, set_routine, cls, msg) \
    (ext_name[0] << 24) + (ext_name[1] << 16) +  (ext_name[2] << 8) + ext_name[3],

// This contains the names of all of the configuration variables names.
static const uint32_t keyer_config_names[] = {

    CONFIGURATION

    0
};

#undef CV

#define CV(v_name, def_val, ext_name, string_name, set_routine, cls, msg) def_val,

// The default configuration struct has the default values for all of
// the configuration variables.
static const config_data keyer_config_default_data = {

  CONFIGURATION
  
};

#undef CV

#define CV(v_name, def_val, ext_name, string_name, set_routine, cls, msg) set_routine,

// The configuration types used to allow changing values from the encoders
static const config_types config_type[] = {

    CONFIGURATION

};

#undef CV

#define CV(v_name, def_val, ext_name, string_name, set_routine, cls, msg) string_name,

// The strings for the status keyer to send when setting configuration parameters
static const char *string_names[] = {

    CONFIGURATION

};

#undef CV

static uint32_t ELEMENTS = sizeof(string_names) / sizeof (char *);


#define CV(v_name, def_val, ext_name, string_name, set_routine, cls, msg) \
    MESSAGE(cls, msg),

static const uint32_t messages[] = {

    CONFIGURATION

};

#undef CV

keyer_config::keyer_config()
{
    // Set up the config values early.  Lots of other code
    // depends on them.
    restore();
}

void
keyer_config::setup()
//----------------------------------------------------------------------
// Set up the configuration instance
//----------------------------------------------------------------------
{
    queue.subscribe(CLASS_CONFIG, this);

    // Select the initial parameter for setup mode.  The parameters
    // are defined so that turning the encoder counterclockwise one
    // click will bring it to the defaults setting.  This is deliberate.
    // It provides a way to re-acquire defaults if something has been
    // messed up.
    for (uint32_t i=0; i<ELEMENTS; i++) {
        if (config_type[i] != T_NONE) {
            m_param = i;
            break;
        }
    }
}

uint32_t
keyer_config::get_handle(uint8_t *name)
//----------------------------------------------------------------------
// Get the handle (actually offset) of a configuration parameter
//----------------------------------------------------------------------
{
    uint32_t vname = (name[0] << 24) + (name[1] << 16) +  (name[2] << 8) + name[3];
    
    for (uint32_t i=0; keyer_config_names[i]; i++) {
        if (vname == keyer_config_names[i]) {
            return i;
        }
    }

    // The configuration parameter does not exist.
    return NO_CONFIGURATION;
}

uint8_t
keyer_config::get_by_handle(uint32_t handle)
//----------------------------------------------------------------------
// Get the value of a configuration parameter specified by a handle
//----------------------------------------------------------------------
{
    uint8_t *p = (uint8_t *)&m_data;
    return p[handle];
}

void
keyer_config::get_name_by_handle(uint32_t handle, uint8_t *buffer)
//----------------------------------------------------------------------
// Get the name of a configuration parameter specified by a handle
//----------------------------------------------------------------------
{
    uint32_t name = keyer_config_names[handle];
    buffer[0] = name >> 24;
    buffer[1] = (name >> 16) & 0xFF;
    buffer[2] = (name >> 8) & 0xFF;
    buffer[3] = name & 0xFF;
    buffer[4] = '\0';    
}
    
void
keyer_config::set_by_handle(uint32_t handle, uint8_t val)
//----------------------------------------------------------------------
// Set the value of a configuration parameter specified by a handle
//----------------------------------------------------------------------
{
    uint8_t *p = (uint8_t *)&m_data;
    p[handle] = val;
    if (messages[handle]) {
        queue.put(messages[handle], val);
    }
}

void
keyer_config::save()
//----------------------------------------------------------------------
// Write the configuration into flash memory
//----------------------------------------------------------------------
{
    union {
        config_data *ptr;
        uint32_t val;
    } flash_data;

    bool erase = false;
    
    flash_data.ptr = get_last_good_flash();
    if (flash_data.val) {
        flash_data.val += FLASH_DATA_SIZE;
        if ((flash_data.val + FLASH_DATA_SIZE) > FLASH_PAGE_END) {
            // The page is full.
            erase = true;
        }
    }
    else {
        flash_data.val = FLASH_PAGE;
    }

    // Verify that the area to be written is indeed erased
    char *cp = (char *)flash_data.ptr;
    for (uint32_t i=0; i<FLASH_DATA_SIZE; i++) {
        if (cp[i] != 0xff) {
            erase = true;
            break;
        }
    }

    FLASH_Unlock();

    if (erase) {
        // Erase the and start at the beginning
        if (FLASH_ErasePage(FLASH_PAGE) != FLASH_COMPLETE) {
            error.general("Flash erase failed");
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "SRI");
            FLASH_Lock();
            return;
        }
        flash_data.val = FLASH_PAGE;
    }

    // Write the flash data
    uint16_t* source = (uint16_t *)&m_data;
    uint32_t dest = flash_data.val;
    for (uint32_t i=0; i<(FLASH_DATA_SIZE / 2); i++) {
        if (FLASH_ProgramHalfWord(dest, source[i]) != FLASH_COMPLETE) {
            // Erase the page rather than leave a partially written configuration
            FLASH_ErasePage(FLASH_PAGE);
            error.general("Flash write failed");
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "SRI");
            break;
        }
        dest += 2;
    }

   FLASH_Lock();
}

void
keyer_config::restore()
//----------------------------------------------------------------------
// Read the configuration from flash memory
//----------------------------------------------------------------------
{
    // If the configuration has been saved to flash get it from there.
    // Otherwise use the default values.
    config_data *flash_data = get_last_good_flash();
    if (flash_data) {
        memcpy(&m_data, flash_data, sizeof(m_data));
    }
    else {
        memcpy(&m_data, &keyer_config_default_data, sizeof(m_data));
    }
}

void
keyer_config::notify()
//----------------------------------------------------------------------
// Notify an object that a parameter has been changed
//----------------------------------------------------------------------
{
    uint8_t *p = (uint8_t *)&m_data;
    for (uint32_t i=0; i<ELEMENTS; i++) {
        if (messages[i]) {
            queue.put(messages[i], p[i]);
        }
    }
}

config_data *
keyer_config::get_last_good_flash()
//----------------------------------------------------------------------
// Find the most recent configuration stored in flash memory
//----------------------------------------------------------------------
{
    config_data *flash_data = NULL;
    union {
        config_data *ptr;
        uint32_t val;
    } next_data;

    next_data.val = FLASH_PAGE;
    while ((next_data.ptr->first == 12) && (next_data.ptr->last == 34)) {
        flash_data = next_data.ptr;

        uint32_t next = next_data.val + FLASH_DATA_SIZE;
        if ((next + FLASH_DATA_SIZE) > FLASH_PAGE_END) {
            break;
        }

        next_data.val = next;
    }

    return flash_data;
}

#ifdef DEBUG
void
keyer_config::display()
{
//----------------------------------------------------------------------
// Display all configuration values.  For debugging only
//----------------------------------------------------------------------
#define CV(v_name, def_val, ext_name, string_name, set_routine, cls, msg) \
    Serial.print( #v_name ": "); Serial.println(m_data.v_name, DEC);

    CONFIGURATION

#undef CV    
}
#endif

void
keyer_config::notify(uint32_t param)
//----------------------------------------------------------------------
// Notify the PC protocol handler that a parameter has changed
//----------------------------------------------------------------------
{
    queue.put(MESSAGE(CLASS_PROTOCOL, MSG_CONFIG_CHANGED), param);
}

void
keyer_config::change_value(uint32_t change)
//----------------------------------------------------------------------
// Change a configuration parameter because of an encoder action
//----------------------------------------------------------------------
{
    config_types ct = config_type[m_param];
    uint8_t *p = (uint8_t *)&m_data;
    uint8_t val = p[m_param];
            
    switch(change) {
        case CHANGE_UP:
        {
            switch (ct) {
                case T_SPEED:
                    queue.put(MESSAGE(CLASS_STATUS, MSG_SPEED_UP), 0);
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_DIT), 0);
                    break;

                case T_FREQ:
                    queue.put(MESSAGE(CLASS_BEEP, MSG_FREQUENCY_UP), OWNER_STATUS);
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_DIT), 0);
                    break;

                case T_VOL:
                    queue.put(MESSAGE(CLASS_BEEP, MSG_VOLUME_UP), OWNER_STATUS);
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_DIT), 0);
                    break;
                    
                case T_WEIGHT:
                    val++;
                    if (val > KEYER_MAX_WEIGHT) {
                        val = KEYER_MAX_WEIGHT;
                    }
                    p[m_param] = val;
                    break;

                case T_0255:
                    if (val < 255) {
                        val++;
                        p[m_param] = val;
                    }
                    break;

                case T_ONOFF:
                    if (val) {
                        val = 0;
                        queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "OFF");
                    }
                    else {
                        val = 1;
                        queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "ON");
                    }
                    p[m_param] = val;

                    break;

                default:
                    break;
            }
            break;        
        }

        case CHANGE_DOWN:
        {
            switch (ct) {
                case T_SPEED:
                    queue.put(MESSAGE(CLASS_STATUS, MSG_SPEED_DOWN), 0);
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_DIT), 0);
                    break;

                case T_FREQ:
                    queue.put(MESSAGE(CLASS_BEEP, MSG_FREQUENCY_DOWN), OWNER_STATUS);
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_DIT), 0);
                    break;

                case T_VOL:
                    queue.put(MESSAGE(CLASS_BEEP, MSG_VOLUME_DOWN), OWNER_STATUS);
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_DIT), 0);
                    break;
                    
                case T_WEIGHT:
                    val++;
                    if (val < KEYER_MIN_WEIGHT) {
                        val = KEYER_MIN_WEIGHT;
                    }
                    p[m_param] = val;
                    break;

                case T_0255:
                    if (val > 0) {
                        val--;
                        p[m_param] = val;
                    }
                    break;

                case T_ONOFF:
                    if (val) {
                        val = 0;
                        queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "OFF");
                    }
                    else {
                        val = 1;
                        queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "ON");
                    }
                    p[m_param] = val;
                    
                    break;

                default:
                    break;
            }
            break;        
        }

        case CHANGE_NONE:
            break;
    }
}

void
keyer_config::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle configuration messages
//----------------------------------------------------------------------
{
    switch (MESSAGE_ID(message)) {
        case MSG_CONFIG_START:
        {
            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "SETUP");
            break;
        }

        case MSG_PARAM_CHANGE:
        {
            if (data == CHANGE_UP) {
                do {
                    m_param++;
                    if (m_param == ELEMENTS) {
                        m_param = 0;
                    }
                } while(config_type[m_param] == T_NONE);

            }
            else {
                do {
                    if (m_param == 0) {
                        m_param = ELEMENTS;
                    }
                    m_param--;
                } while(config_type[m_param] == T_NONE);
            }

            queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), string_names[m_param]);
            break;
        }

        case MSG_VALUE_CHANGE:
        {
            change_value(data);
            break;
        }

        case MSG_CONFIG_SET:
        {
            config_types ct = config_type[m_param];
            uint8_t *p = (uint8_t *)&m_data;
            uint32_t val = p[m_param];

            if (m_data.save) {
                m_data.save = 0;
                save();
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "R");
                break;
            }

            if (m_data.restore) {
                restore();
                notify();
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "R");
                break;
            }

            if (m_data.defaults) {
                memcpy(&m_data, &keyer_config_default_data, sizeof(m_data));
                notify();
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "R");
                break;
            }

            switch (ct) {
                case T_FREQ:
                    queue.put(MESSAGE(CLASS_BEEP, MSG_VALUES_XFREQ), 0);
                    break;

                case T_VOL:
                    queue.put(MESSAGE(CLASS_BEEP, MSG_VALUES_XVOL), 0);
                    break;

                case T_SPEED:                
                case T_WEIGHT:
                case T_0255:
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_CLEAR), 0);
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_NUMBER), val);
                    break;

                case T_ONOFF:
                    if (val) {
                        queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "ON");
                    }
                    else {
                        queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "OFF");
                    }
                    break;

                default:
                    break;
            }
        }

        default:
            error.bad_message("Config", message, data);

    }
}

keyer_config config;
