/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include "keyer_pins.h"
#include "keyer_config.h"
#include "keyer_control.h"
#include "keyer_encoder.h"
#include "keyer_error.h"
#include "keyer_messages.h"
#include "keyer_misc.h"
#include "keyer_queue.h"

keyer_control::keyer_control() :
    m_encoder1_pressed(false), m_encoder2_pressed(false),
    m_encoder1_used(false), m_encoder2_used(false)
{
  
}

// Handle the front panel encoder controls.  There are two encoders,
// each with a pushbutton.  This produces lots of combinations which
// can do different things.
//
//
// In normal mode:
//
//   Rotate encoder 1 - change keyer speed
//
//   Press and rotate encoder 1 - change keyer type
//
//   Press and release encoder 1 - send keyer speed using the status
//   keyer
//
//   Press and hold encoder 1 - reverse paddle
//
//   Rotate encoder 2 - change sidetone volume
//
//   Press and rotate encoder 2 - change sidetoe frequency
//
//   Press and release encoder 2 - send sidetone volume and frequency
//   using the status keyer
//
//   Press and hold encoder 2 - Switch between speaker and headphones
//
//   Press encoder 1 and encoder 2 - enter setup mode
//
// In setup mode:
//
//   Rotate encoder 1 - change to next or previous setup parameter
//
//   Press encoder 1 - exit setup mode, enter normal mode
//
//   Rotate encoder 2 - change parameter value
//
//   Press encoder 2 - Save parameter or take action based on setting

// This class is hardware dependent only in the setup routine which
// initializes the encoder instances with the I/O pin numbers they are
// connected to.  That code is dependent on the I/O structure of the
// processor.

void keyer_control::setup()
//----------------------------------------------------------------------
// Set up for control (encoder or button) changes
//----------------------------------------------------------------------
{
    queue.subscribe(CLASS_CONTROL, this);

    m_encoder1_timer = queue.get_timer();
    m_encoder2_timer = queue.get_timer();

    encoder1.setup(CLASS_ENC1, ENC1_A, ENC1_B, ENC1_S);
    encoder2.setup(CLASS_ENC2, ENC2_A, ENC2_B, ENC2_S);

    m_setup_mode = false;
}

void keyer_control::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle messages from the encoders and buttons
//----------------------------------------------------------------------
{
    switch (MESSAGE_ID(message)) {
    
        case MSG_BUTTON_DOWN:
        {
            if (data == CLASS_ENC1) {
                m_encoder1_pressed = true;
                m_encoder1_used = false;

                // Pressing encoder 1 ends setup mode
                if (m_setup_mode) {
                    m_setup_mode = false;
                    m_encoder1_used = true;
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "OK");
                    break;
                }

                // Pressing encoder 1 and encoder 2 enters setup mode
                if (m_encoder2_pressed) {
                    m_setup_mode = true;
                    m_encoder1_used = true;
                    m_encoder2_used = true;
                    queue.put(MESSAGE(CLASS_CONFIG, MSG_CONFIG_START), 0);
                    break;
                }
                queue.put_timed(m_encoder1_timer, config.get_long_press() * TIMER_PER_MS * 100,
                                MESSAGE(CLASS_CONTROL, MSG_LONG_PRESS), 1);
            }
            else {
                m_encoder2_pressed = true;
                m_encoder2_used = false;

                // Pressing encoder 2 sets a value
                if (m_setup_mode) {
                    m_encoder2_used = true;
                    queue.put(MESSAGE(CLASS_CONFIG, MSG_CONFIG_SET), 0);
                    break;
                }

                // Pressing encoder 1 and encoder 2 enters setup mode
                if (m_encoder1_pressed) {
                    m_setup_mode = true;
                    m_encoder1_used = true;
                    m_encoder2_used = true;
                    queue.put(MESSAGE(CLASS_CONFIG, MSG_CONFIG_START), 0);
                    break;
                }

                queue.put_timed(m_encoder2_timer, config.get_long_press() * TIMER_PER_MS * 100,
                                MESSAGE(CLASS_CONTROL, MSG_LONG_PRESS), 2);
            }
            break;
        }

        case MSG_BUTTON_UP:
        {
            if (data == CLASS_ENC1) {
                queue.cancel_timed(m_encoder1_timer);
                if (!m_encoder1_used) {
                    // Press-release is send speed as status
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_CLEAR), 0);
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_NUMBER), config.get_keyer_speed());
                }
                m_encoder1_pressed = false;
            }
            else {
                queue.cancel_timed(m_encoder2_timer);
                if (!m_encoder2_used) {
                    // Press-release is send sidetone status
                    queue.put(MESSAGE(CLASS_BEEP, MSG_VALUES), 0);
                }
                m_encoder2_pressed = false;
            }
            break;
        }

        case MSG_ENCODER_CCW:
        {
            if (data == CLASS_ENC1) {
                m_encoder1_used = true;

                if (m_setup_mode) {
                    queue.put(MESSAGE(CLASS_CONFIG, MSG_PARAM_CHANGE), CHANGE_DOWN);
                    break;
                }

                if (m_encoder1_pressed) {
                    // Keyer type
                    queue.put(MESSAGE(CLASS_KEYER, MSG_TYPE_DOWN), 0);
                }
                else {
                    // Speed
                    queue.put(MESSAGE(CLASS_KEYER, MSG_SPEED_DOWN), 0);
                }
            }
            else {  // Encoder 2
                m_encoder2_used = true;

                if (m_setup_mode) {
                    queue.put(MESSAGE(CLASS_CONFIG, MSG_VALUE_CHANGE), CHANGE_DOWN);
                    break;
                }

                if (m_encoder2_pressed) {
                    //  Volume
                    queue.put(MESSAGE(CLASS_BEEP, MSG_FREQUENCY_DOWN), OWNER_KEYER);
                }
                else {
                    // Frequency
                    queue.put(MESSAGE(CLASS_BEEP, MSG_VOLUME_DOWN), OWNER_KEYER);
                }
            }
            break;
        }

        case MSG_ENCODER_CW:
        {
            if (data == CLASS_ENC1) {
                m_encoder1_used = true;

                if (m_setup_mode) {
                    queue.put(MESSAGE(CLASS_CONFIG, MSG_PARAM_CHANGE), CHANGE_UP);
                    break;
                }

                if (m_encoder1_pressed) {
                    // Keyer type
                    queue.put(MESSAGE(CLASS_KEYER, MSG_TYPE_UP), 0);
                }
                else {
                    // Speed
                    queue.put(MESSAGE(CLASS_KEYER, MSG_SPEED_UP), 0);
                }
            }
            else {  // Encoder 2
                m_encoder2_used = true;

                if (m_setup_mode) {
                    queue.put(MESSAGE(CLASS_CONFIG, MSG_VALUE_CHANGE), CHANGE_UP);
                    break;
                }

                if (m_encoder2_pressed) {
                    //  Volume
                    queue.put(MESSAGE(CLASS_BEEP, MSG_FREQUENCY_UP), OWNER_KEYER);
                }
                else {
                    // Frequency
                    queue.put(MESSAGE(CLASS_BEEP, MSG_VOLUME_UP), OWNER_KEYER);
                }
            }
            break;
        }

        case MSG_LONG_PRESS:
        {
            if (data == 1) {
                if (m_encoder1_used || !m_encoder1_pressed) {
                    break;
                }
              
                m_encoder1_used = true;
                // Press and hold is paddle swap
                queue.put(MESSAGE(CLASS_KEYER, MSG_PADDLE_SWAP), 0);
            }
            else {
                if (m_encoder2_used || !m_encoder2_pressed) {
                    break;
                }
              
                m_encoder2_used = true;
                // Press and hold is headphones/speaker swap
                uint8_t use_phones = config.get_use_phones();
                if (use_phones) {
                    config.set_use_phones(0);
                }
                else {
                    config.set_use_phones(1);
                }

                queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_UPDATE), OWNER_KEYER);

                if (use_phones) {
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "SPKR");
                }
                else {
                    queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "PHONES");
                }
            }

            break;
        }

        default:
            error.bad_message("Control", message, data);
    }
}

keyer_control control;
