/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include "keyer_pins.h"
#include "keyer_config.h"
#include "keyer_encoder.h"
#include "keyer_error.h"
#include "keyer_messages.h"
#include "keyer_misc.h"
#include "keyer_queue.h"

// This class handles the hardware-dependent interfacing to the encoders.
//
// The KY-040 shows a full quadrature cycle for each click.  Watch for the
// A output going high and then check the B output to determine the direction.
//
// It also has a sloppy quadrature and it bounces.  This code takes an
// interrupt when the encoder changes.  The interrupt handler queues a
// message which will be delivered a few milliseconds later.  When the
// message is delivered the encoder is read.  This isn't perfect and it
// may be possible to improve the handling of the encoder when it is
// turned rapidly.
//
// This class is hardware dependent.  It depends on the I/O library
// routines which in turn are dependent on the processor I/O subsystem.
// The Arduino libraries abstract some of it but the interrupt
// processing is processor-specific.
    
keyer_encoder::keyer_encoder()
    : m_encoder_debounce(false), m_button_debounce(false)
{

}

void
keyer_encoder::setup(uint32_t id, uint8_t pin_a, uint8_t pin_b,
                     uint8_t pin_s)
//----------------------------------------------------------------------
// Set up the specified encoder
//----------------------------------------------------------------------
{
    m_id = id;

    m_pin_a = pin_a;
    m_pin_b = pin_b;
    m_pin_s = pin_s;
    
    // The encoder pins have pull-up resistors.  The switch does not.
    pinMode(m_pin_a, INPUT);
    pinMode(m_pin_b, INPUT);
    pinMode(m_pin_s, INPUT_PULLUP);

    queue.subscribe(m_id, this);

    m_encoder_timer = queue.get_timer();
    m_button_timer = queue.get_timer();
    
    m_button = digitalRead(m_pin_s);

    // The current algorithm doesn't use an interrupt on encoder B but
    // the pin arrangement allows that interrupt so feel free to use
    // that interrupt if a better encoder algorithm requires it.
    attachInterrupt(m_pin_a, keyer_encoder::encoder_interrupt, this, RISING);
    attachInterrupt(m_pin_s, keyer_encoder::button_interrupt, this, CHANGE);
}

void
keyer_encoder::encoder_interrupt(void *data)
//----------------------------------------------------------------------
// Rotary encoder interrupt handler
//----------------------------------------------------------------------
{
    keyer_encoder *e = (keyer_encoder *)data;
    if (!e->m_encoder_debounce) {
        queue.put(MESSAGE(e->m_id, MSG_ENCODER_INTERRUPT), 
                          digitalRead(e->m_pin_b));
        e->m_encoder_debounce = true;
    }
}

void
keyer_encoder::button_interrupt(void *data)
//----------------------------------------------------------------------
// Encoder push-button interrupt handler
//----------------------------------------------------------------------
{
    keyer_encoder *e = (keyer_encoder *)data;
    if (!e->m_button_debounce) {
        queue.put_timed(e->m_button_timer, config.get_button_debounce()
                        * TIMER_PER_MS,
                        MESSAGE(e->m_id, MSG_BUTTON_DEBOUNCE), 0);
        e->m_button_debounce = true;
    }
}

void
keyer_encoder::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle encoder rotary and push-button messages
//----------------------------------------------------------------------
{  
    switch (MESSAGE_ID(message)) {
        case MSG_ENCODER_INTERRUPT:
        {
            // The output of the encoder is dirty.  This could be a glitch.
            if (!digitalRead(m_pin_a)) {
                m_encoder_debounce = false;
                return;
            }

            m_b_val = digitalRead(m_pin_b);

            queue.put_timed(m_encoder_timer, config.get_encoder_debounce()
                            * TIMER_PER_MS,
                            MESSAGE(m_id, MSG_ENCODER_DEBOUNCE), 0);
            break;
        }

        case MSG_ENCODER_DEBOUNCE:
        {
            m_encoder_debounce = false;

            // Check that pin A is still high.  If not this was a glitch.
            if (!digitalRead(m_pin_a)) {
                return;
            }

            // Pin B can also glitch.  Consider it low if it was low at the
            // time of the interrupt or now.
            if (m_b_val && digitalRead(m_pin_b)) {
                queue.put(MESSAGE(CLASS_CONTROL, MSG_ENCODER_CCW), m_id);
            }
            else {
                queue.put(MESSAGE(CLASS_CONTROL, MSG_ENCODER_CW), m_id);
            }

            break;
        }

        case MSG_BUTTON_DEBOUNCE:
        {
            m_button_debounce = false;

            // Button is connected to ground - 0 is pressed
            if (digitalRead(m_pin_s)) {
                queue.put(MESSAGE(CLASS_CONTROL, MSG_BUTTON_UP), m_id);
            }
            else {
                queue.put(MESSAGE(CLASS_CONTROL, MSG_BUTTON_DOWN), m_id);
            }

            m_button = data;
            
            break;
        }

        default:
            error.bad_message("Encoder", message, data);
    }
}

keyer_encoder encoder1;
keyer_encoder encoder2;
