/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include <wirish.h>

#include "keyer_error.h"
#include "keyer_keyer.h"
#include "keyer_queue.h"
#include "keyer_iambic.h"

// This provides the manual (paddle) control for the keyer in cases
// where the keyer clock is asynchronous and it has memories.
// The code in keyer_keyer calls this through the public routines
// inherited from keyer_manual.  And this code calls the keyer through
// the public routines in that code.
//
// This class is hardware independent.

keyer_iambic::keyer_iambic()
{
    
}

bool
keyer_iambic::setup(keyer_type type, bool silent)
//----------------------------------------------------------------------
// If the keyer type requested is one of the supported types set it up
//----------------------------------------------------------------------
{
    switch (type) {
        case KEYER_TYPE_IAMBIC_A:
            m_type_b = false;
            if (!silent) {
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "IA");
            }
            break;

        case KEYER_TYPE_IAMBIC_B:
            m_type_b = true;
            if (!silent) {
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "IB");
            }
            break;

        case KEYER_TYPE_ULTIMATIC:
            m_type_b = config.get_keyer_default_b();
            if (!silent) {
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "ULT");
            }
            break;

        case KEYER_TYPE_DIT:
            m_type_b = config.get_keyer_default_b();
            if (!silent) {
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "DIT");
            }
            break;

        case KEYER_TYPE_DAH:
            m_type_b = config.get_keyer_default_b();
            if (!silent) {
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "DAH");
            }
            break;

        default:
            return false;

    }

    m_cap_left = false;
    m_cap_right = false;
    m_phy_left = false;
    m_phy_right = false;

    m_both = false;
    clear_latch();

    m_idle = true;

    keyer.km_set_echo(true);
    return true;
}

void keyer_iambic::paddle_change(uint32_t id, uint32_t data)
//----------------------------------------------------------------------
// Handle a paddle opening or closing
//----------------------------------------------------------------------
{
    bool left_on = false;
    bool right_on = false;
    
    switch (id) {
        case MSG_CAP_LEFT_CHANGED:
        {
            if (data) {
                m_cap_left = true;
                m_left_pressed_last = true;
                left_on = true;
            }
            else {
                m_cap_left = false;
            }
    
            break;
        }

        case MSG_CAP_RIGHT_CHANGED:
        {
            if (data) {
                m_cap_right = true;
                m_left_pressed_last = false;
                right_on = true;
            }
            else {
                m_cap_right = false;
            }
    
            break;
        }

        case MSG_PHY_LEFT_CHANGED:
        {
            if (data) {
                m_phy_left = true;
                m_left_pressed_last = true;
                left_on = true;
            }
            else {
                m_phy_left = false;
            }
    
            break;
        }
 
        case MSG_PHY_RIGHT_CHANGED:
        {
            if (data) {
                m_phy_right = true;
                m_left_pressed_last = false;
                right_on = true;
            }
            else {
                m_phy_right = false;
            }
    
            break;
        }
    }

    if (m_idle) {
        if (left_on) {
            send_left();
            m_idle = false;
        }
        else {
            if (right_on) {
                send_right();
                m_idle = false;
            }
        }
    }

    check_latch();
}

void
keyer_iambic::state_change(keyer_state state)
//----------------------------------------------------------------------
// Handle a keyer state change
//----------------------------------------------------------------------
{
    switch (state) {
        case KEYER_STATE_CHAR_SPACE:
        {
            break;
        }

        case KEYER_STATE_IDLE:
        {
            if (left_paddle()) {
                m_left_seen = true;
            }
            if (right_paddle()) {
                m_right_seen = true;
            }

            // Iambic B sends an extra dit or dah if both were pressed.  The behavior of
            // Ultimatic, dit, and dah can be set to A or B style by a configuration parameter.
           if (!m_type_b) {
                if (m_both && !left_paddle() && !right_paddle()) {
                    clear_latch();
                }
            }
            m_both = false;

            if (m_left_seen) {
                if (m_right_seen) {
                    switch (config.get_keyer_type()) {
                        case KEYER_TYPE_IAMBIC_A:
                        case KEYER_TYPE_IAMBIC_B:
                            if (m_left_sent_last) {
                                send_right();
                            }
                            else {
                                send_left();
                            }
                            break;

                        case KEYER_TYPE_ULTIMATIC:
                            if (m_left_pressed_last) {
                                send_left();
                            }
                            else {
                                send_right();
                            }
                            break;

                        case KEYER_TYPE_DIT:
                            km_send_dit();
                            break;

                        case KEYER_TYPE_DAH:
                            km_send_dah();
                            break;

                        default:
                            error.bad_state("state_change", state, 0);
                    }
                    m_both = true;
                }
                else {
                    send_left();
                }
            }
            else {
                if (m_right_seen) {
                    send_right();
                }
                else {
                    clear_latch();
                    if (config.get_keyer_autospace()) {
                        // Open the memory latch.  This makes it easier to send
                        // with autospace.
                        m_left_ok = true;
                        m_right_ok = true;
                        keyer.km_send_space();
                    }
                    else {
                        m_idle = true;
                    }
                }
            }
            break;
        }

        default:
            break;
    }
}

keyer_iambic iambic;
