/******************************************************************************
 * The MIT License
 *s
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include "keyer_common.h"
#include "keyer_error.h"
#include "keyer_keyer.h"
#include "keyer_messages.h"
#include "keyer_misc.h"
#include "keyer_protocol.h"
#include "keyer_queue.h"

// The keyer accepts input from the PC through the PC or manually through
// a capacitive or physical paddle.  The protocol code for the PC and the
// manual key code call public routines in the keyer.
//
// This class is hardware independent.

// These are all of the known manual keyers
#include "keyer_bug.h"
#include "keyer_iambic.h"
#include "keyer_straight_key.h"
#include "keyer_ultimatic.h"

static keyer_manual *manual_keyers[] = {&iambic, &ultimatic, &straight_key, &bug, NULL};

#define KEY_DOWN                                                        \
    queue.put(MESSAGE(CLASS_OUTPUTS, MSG_OUTPUT_CHANGED), OUTPUT_ON);   \
    queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_ON), OWNER_KEYER)

#define KEY_UP                                                          \
    queue.put(MESSAGE(CLASS_OUTPUTS, MSG_OUTPUT_CHANGED), OUTPUT_OFF);  \
    queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_OFF), OWNER_KEYER)

#define PTT_ON                                                          \
    m_ptt = true;                                                       \
    queue.put(MESSAGE(CLASS_OUTPUTS, MSG_PTT_CHANGED), OUTPUT_ON)

#define PTT_OFF                                                         \
    m_ptt = false;                                                      \
    queue.put(MESSAGE(CLASS_OUTPUTS, MSG_PTT_CHANGED), OUTPUT_OFF)

keyer_keyer::keyer_keyer()
{
    m_sequence = 0;
    m_ptt_sequence = 0;
    m_deadman_sequence = 0;
    m_sync_idles = 0xff;
    m_ptt = false;

    // Fill in the reverse Morse table which is used to convert
    // what is sent with the paddle to ASCII.
    memset (m_reverse_morse, 0, sizeof(m_reverse_morse));
    for (uint32_t i=0; i<sizeof(morse_char); i++) {
        m_reverse_morse[morse_char[i]] = i + 32;
    }
}

void
keyer_keyer::setup()
//----------------------------------------------------------------------
// Set up the main keyer
//----------------------------------------------------------------------
{
    queue.subscribe(CLASS_KEYER, this);
 
    m_timer = queue.get_timer();
    m_ptt_timer = queue.get_timer();
    m_deadman_timer = queue.get_timer();

    m_state = KEYER_STATE_IDLE;
    m_paddle = true;

    uint8_t speed = config.get_keyer_encoder();
    if (speed > KEYER_MAX_SPEED) {
        speed = KEYER_MAX_SPEED;
    }
    if (speed < KEYER_MIN_SPEED) {
        speed = KEYER_MIN_SPEED;
    }
    config.set_keyer_encoder(speed);
    protocol.kr_pot_speed(speed);

    m_alter_speed = 0;
    m_alter_weight = 0;
    m_computer_speed = speed;
    m_computer_sets_speed = false;
    
    update_speed();

    m_sent = 0;
    m_sent_mask = 1;

    m_type = KEYER_TYPE_FENCE;
    set_keyer_type((keyer_type)config.get_keyer_type(), true);
}

void
keyer_keyer::update_speed()
//----------------------------------------------------------------------
// Set the keyer speed and weight
//----------------------------------------------------------------------
{
    uint32_t speed;

    uint32_t weight = config.get_keyer_weight();
    if (m_computer_sets_speed) {
        speed = m_computer_speed;
    }
    else {
        speed = config.get_keyer_encoder();
    }
    config.set_keyer_speed(speed);

    // Apply paddle alterations or computer delta
    if (m_paddle) {
        speed += (speed * m_alter_speed) / 100;
        weight += (weight * m_alter_weight) / 100;
    }

    if (speed < KEYER_MIN_SPEED) {
        speed = KEYER_MIN_SPEED;
    }
    if (speed > KEYER_MAX_SPEED) {
        speed = KEYER_MAX_SPEED;
    }

    if (weight < KEYER_MIN_WEIGHT) {
        weight = KEYER_MIN_WEIGHT;
    }

    if (weight > KEYER_MAX_WEIGHT) {
        weight = KEYER_MAX_WEIGHT;
    }

    // The weight calculation affects the actual speed because at weights other than
    // 50% the length of a dah plus the key up after it is not twice the length of a
    // dit plus the key up after it.  This speed difference is ignored.  If this is
    // a problem it could be compensated for in a way similar to what is done for
    // Farnsworth.
    m_base_ticks = TIMER_UNIT / speed;
    m_dit_down_ticks = (m_base_ticks * weight) / 50;
    m_dah_down_ticks = m_dit_down_ticks * 3;
    m_up_ticks = (m_base_ticks * 2) - m_dit_down_ticks;
    m_char_ticks = CHAR_TICKS * m_base_ticks;
        

    // Farnsworth is in effect if the Farnsworth speed is greater than the keyer speed.
    // This can have weird effects if the computer sends delta speeds.  Farnsworth is
    // only used when sending from the computer, never from the paddle.
    if (!m_paddle && (config.get_keyer_farnsworth() > speed)) {

        // This follows the algorithm in April 1990 QEX, A Standard for Morse Timing
        // Using the Farnsworth Technique by Jon Bloom, KE3Z, ARRL Laboratory.
        //
        // Normally a unit, which is the key-down part of a dot, is 1.2/(speed in WPM)
        // seconds.  Dah is 3 units, character space is 3 units, word space is 7 and
        // units are all the same length.
        //
        // In Farnsworth the units for dit and dah are set by the Farnsworth speed and
        // the character and word spaces are made longer so that the result comes out
        // at the set speed.
        //
        // Some keyers apparently just use the keyer speed for the character and word
        // space units but this results in a higher speed than what was set because
        // the dit and dah times are short.  The character and word spaces must be
        // made longer to compensate for the short dits and dahs.
        //
        // The compensation is based on the standard word PARIS which has 50 units.
        // Of these units 31 are dits and dahs and 19 are character and word spaces.
        //
        // Assume s = speed, c = character speed (Farnsworth speed).
        //
        // Sending PARIS should take 50 * 1.2/s seconds.  Sending the dits and dahs
        // should take 31 * 1.2/c seconds.  So the character and word spaces should
        // take the difference between those times.
        //
        //         (50 * 1.2)    (31 * 1.2)
        //    ta = ---------- -  ----------
        //             s             c
        //
        // There are 19 character and word space units.  So each space unit is ta/19.
        //
        m_computer_base_ticks = TIMER_UNIT / config.get_keyer_farnsworth();
        m_computer_dit_down_ticks = (m_computer_base_ticks * config.get_keyer_weight()) / 50;
        m_computer_dah_down_ticks = m_computer_dit_down_ticks * 3;
        m_computer_up_ticks = (m_computer_base_ticks * 2) - m_computer_dit_down_ticks;

        // The keyer sends each dit or dah with the corresponding key-up.  Subtract that
        // from the character and word spaces.  Use the base ticks value for this because
        // it should be constant no matter what the weight.
        uint32_t space_unit = ((50 * m_base_ticks) - (31 * m_computer_base_ticks)) / 19;
        m_computer_char_ticks = (CHAR_TICKS * space_unit) - m_base_ticks;
        if (config.get_keyer_space_6()) {
            m_computer_word_ticks = ((WORD_TICKS_SHORT - CHAR_TICKS - 1) * space_unit)
                                    - m_base_ticks;
        }
        else {
            m_computer_word_ticks = ((WORD_TICKS_LONG - CHAR_TICKS - 1) * space_unit)
                                    - m_base_ticks;
        }
    }
    else {
        m_computer_base_ticks = m_base_ticks;
        m_computer_dit_down_ticks = m_dit_down_ticks;
        m_computer_dah_down_ticks = m_dah_down_ticks;
        m_computer_up_ticks = m_up_ticks;
        m_computer_char_ticks = (CHAR_TICKS * m_base_ticks) - m_base_ticks;
        if (config.get_keyer_space_6()) {
            m_computer_word_ticks = ((WORD_TICKS_SHORT - CHAR_TICKS - 1)* m_base_ticks)
                                    - m_base_ticks;
        }
        else {
            m_computer_word_ticks = ((WORD_TICKS_LONG - CHAR_TICKS - 1) * m_base_ticks)
                                    - m_base_ticks;
        }        
    }
}

void
keyer_keyer::set_ticks(uint32_t ticks)
//----------------------------------------------------------------------
// Start a timed operation for key up or down
//----------------------------------------------------------------------
{
    m_sequence++;
    queue.put_timed(m_timer, ticks, MESSAGE(CLASS_KEYER, MSG_TIMER),
                    m_sequence);
}

void
keyer_keyer::set_ptt_ticks(uint32_t ticks)
//----------------------------------------------------------------------
// Start a timed operation for PTT pre or post
//----------------------------------------------------------------------
{
    m_ptt_sequence++;
    queue.put_timed(m_ptt_timer, ticks, MESSAGE(CLASS_KEYER, MSG_PTT_TIMER),
                    m_ptt_sequence);
}

void keyer_keyer::set_deadman()
//----------------------------------------------------------------------
// Start a timed operation to stop the keyer if it is key-down too long
//----------------------------------------------------------------------
{
    m_deadman_sequence++;
    queue.put_timed(m_deadman_timer, config.get_keyer_dead_man()
                    * (TIMER_PER_MS * 1000),
                    MESSAGE(CLASS_KEYER, MSG_DEADMAN_TIMER),
                            m_deadman_sequence);
}

void
keyer_keyer::set_keyer_type(keyer_type type, bool silent)
//----------------------------------------------------------------------
// Set the keyer type (iambic, dit, dah, bug etc)
//----------------------------------------------------------------------
{
    config.set_keyer_type(type);

    if (type == m_type) {
        return;
    }

    m_echo = false;

    m_type = type;

    for (int i=0; manual_keyers[i]; i++) {
        if (manual_keyers[i]->setup(type, silent)) {
            m_manual = manual_keyers[i];
            return;
        }
    }
}

void keyer_keyer::start_paddle(keyer_state state)
//----------------------------------------------------------------------
// Start sending from the paddle, handle PTT pre if needed
//----------------------------------------------------------------------
{
    // If sending from the computer stop and switch to the paddle
    if (!m_paddle) {
        stop_computer();
        m_paddle = true;
        update_speed();
    }

    m_sync_idles = 0;

    // If idle then start PTT as necessary
    if ((m_state == KEYER_STATE_IDLE) ||
        (m_state == KEYER_STATE_IDLE_SYNC)) {
        // If PTT is on in idle state then the keyer just finished
        // sending something and it is during the PTT post time.
        if (m_ptt) {
            // Invalidate the PTT off timeout
            m_ptt_sequence++;
        }
        else {
            PTT_ON;
            if (config.get_keyer_ptt_pre()) {
                m_state = KEYER_STATE_PTT_PRE;
                m_next_state = state;
                // PTT pre and post are specified in tens of
                // milliseconds
                set_ticks(config.get_keyer_ptt_pre()
                          * (TIMER_PER_MS * 10));
                return;
            }
        }
    }

    // PTT is set so it is OK to start sending.
    send_paddle(state);
}

void keyer_keyer::send_paddle(keyer_state state)
//----------------------------------------------------------------------
// Send a dit, dah, or tune from the paddle
//----------------------------------------------------------------------
{
    switch (state) {
        case KEYER_STATE_DOWN:
            KEY_DOWN;
            set_deadman();
            // A bug would use this to send dahs.  Count each
            // press as a dah.
            m_sent |= m_sent_mask;
            m_sent_mask <<= 1;
            break;
      
        case KEYER_STATE_DOWN_DIT:
            KEY_DOWN;
            set_ticks(m_dit_down_ticks);
            m_sent_mask <<= 1;
            break;
      
        case KEYER_STATE_DOWN_DAH:
            KEY_DOWN;
            set_ticks(m_dah_down_ticks);
            m_sent |= m_sent_mask;
            m_sent_mask <<= 1;
            break;
      
        default:
            error.bad_state("start_paddle", m_state, state);
    }

    m_state = state;
}

void keyer_keyer::start_computer(keyer_state state)
//----------------------------------------------------------------------
// Start sending from the PC, handle PTT pre if needed
//----------------------------------------------------------------------
{
    // If sending from the padle change to sending from the computer
    if (m_paddle) {
        m_paddle = false;
        update_speed();
    }

    // If idle then start PTT as necessary
    if ((m_state == KEYER_STATE_IDLE) || (m_state == KEYER_STATE_IDLE_SYNC)) {
        // If PTT is on in idle state then the keyer just finished sending
        // something and it is during the PTT post time.
        if (m_ptt) {
            // Invalidate the PTT off timeout
            m_ptt_sequence++;
        }
        else {
            PTT_ON;
            if (config.get_keyer_ptt_pre()) {
                m_state = KEYER_STATE_PTT_PRE;
                m_next_state = state;
                // PTT pre and post are specified in tens of milliseconds
                set_ticks(config.get_keyer_ptt_pre() * (TIMER_PER_MS * 10));
                return;
            }
        }
    }

    // PTT is set so it is OK to start sending.
    m_state = state;
    computer_timer();
}

void keyer_keyer::stop_computer()
//----------------------------------------------------------------------
// Immediately stop sending from the computer
//----------------------------------------------------------------------
{
    if (m_paddle) {
        return;
    }

    KEY_UP;

    // If tuning there is a deadman timer.  Stop it.  It may have just
    // fired so make sure it will be ignored.
    if (m_state == KEYER_STATE_TUNE) {
        queue.cancel_timed(m_deadman_timer);
        m_deadman_sequence++;
    }

    send_or_idle_computer();
}

void
keyer_keyer::paddle_timer()
//----------------------------------------------------------------------
// Handle timer expiration when sending from the paddle
//----------------------------------------------------------------------
{
    switch(m_state) {

        // PTT Pre completed, start sending
        case KEYER_STATE_PTT_PRE:
        send_paddle(m_next_state);
        break;
    
        // Dead man timer expired
        case KEYER_STATE_DOWN:
            km_stop_sending();
            break;

        // Key down part of dit or dah completed
        case KEYER_STATE_DOWN_DIT:
        case KEYER_STATE_DOWN_DAH:
            KEY_UP;

            set_ticks(m_up_ticks);
            m_manual->state_change(KEYER_STATE_UP);
            m_state = KEYER_STATE_UP;
            break;

        // Completed dit or dah
        case KEYER_STATE_UP:
            idle_paddle_keyer();
            break;

        // Completed character space
        case KEYER_STATE_CHAR_SPACE:
            idle_paddle_keyer();
            break;

        // Compoeted synchronous idle clock
        case KEYER_STATE_IDLE_SYNC:
            idle_paddle_keyer();
            break;

        // The user didn't send anything for a while after sending a character.
        // Echo a space.
        case KEYER_STATE_IDLE:
            if (m_echo) {
                protocol.kr_echo(' ');
            }
            break;

        default:
            error.bad_state("paddle_timer", m_state, m_next_state);
    }
}

void
keyer_keyer::idle_paddle_keyer()
//----------------------------------------------------------------------
// Handle keyer transition to idle when sending from the paddle
//----------------------------------------------------------------------
{
    // The manual keyer can change the state when it receives a
    // change to PTT post or idle.  Don't do anything if it gets
    // changed.
    keyer_state previous_state = m_state;
    m_state = KEYER_STATE_IDLE;
    m_manual->state_change(KEYER_STATE_IDLE);
    if ((m_state == KEYER_STATE_IDLE) ||
        (m_state == KEYER_STATE_IDLE_SYNC)) {
        if (config.get_keyer_ptt_post()) {
            // PTT pre and post are specified in tens of
            // milliseconds
            set_ptt_ticks(config.get_keyer_ptt_post()
                          * (TIMER_PER_MS * 10));
        }
        else {
            PTT_OFF;
        }

        // Echo the character if this keyer type supports it
        if (m_echo) {
            m_sent |= m_sent_mask;
            if (m_sent_mask && (m_sent < 0xf0)) {
                char c = m_reverse_morse[m_sent];
                if (c) {
                    protocol.kr_echo(c);
                }
            }
            m_sent_mask = 1;
            m_sent = 0;
        }
    }

    // If the keyer doesn't send for a while a space character
    // should be echoed.
    if ((m_state == KEYER_STATE_IDLE) &&
        (previous_state != KEYER_STATE_CHAR_SPACE) &&
        m_echo) {
        set_ticks(m_base_ticks * 4);
    }
}

void
keyer_keyer::send_or_idle_computer()
//----------------------------------------------------------------------
// Send a character from the computer or go idle if no more characters
//----------------------------------------------------------------------
{
    // If there is no character to be sent tell the protocol.  It may
    // have another character.
    if (!m_char) {
        protocol.kr_state(PC_STATE_IDLE);
    }

    if (m_char) {
        m_bits = morse_char[m_char - 32];
        m_char = '\0';

        // Word space may be 6 or 7 bits long
        if (m_bits == 0xf0) {
            set_ticks(m_computer_word_ticks);
            m_state = KEYER_STATE_SPACE;
            return;
        }

        // half-space
        if (m_bits > 0xf0) {
            set_ticks(m_computer_base_ticks * (m_bits - 0xf0));
            m_state = KEYER_STATE_SPACE;
            return;
        }

        // Character - start with dit or dah
        if (m_bits & 1) {
            set_ticks(m_computer_dah_down_ticks);       
            m_state = KEYER_STATE_DOWN_DIT;
        }
        else {
            set_ticks(m_computer_dit_down_ticks);
            m_state = KEYER_STATE_DOWN_DAH;
        }
        KEY_DOWN;
        m_bits >>= 1;
        return;
    }
    
    m_state = KEYER_STATE_IDLE;
    m_paddle = true;
    if (config.get_keyer_ptt_post()) {
        // PTT pre and post are specified in tens of milliseconds
        set_ptt_ticks(config.get_keyer_ptt_post()
                      * (TIMER_PER_MS * 10));
    }
    else {
        PTT_OFF;
    }
}

void
keyer_keyer::computer_timer()
//----------------------------------------------------------------------
// Handle timer expiration when sending from the computer
//----------------------------------------------------------------------
{
    if (m_state == KEYER_STATE_PTT_PRE) {
        m_state = m_next_state;
    }
    
    switch (m_state) {
        // If idle, check for a character to send
        case KEYER_STATE_IDLE:
        {
            send_or_idle_computer();
            break;
        }

        // Key down part of dit completed
        case KEYER_STATE_DOWN_DIT:
        case KEYER_STATE_DOWN_DAH:
        {
            KEY_UP;

            set_ticks(m_computer_up_ticks);
            m_state = KEYER_STATE_UP;
            break;
        }

        // Key up part of dit or day completed
        case KEYER_STATE_UP:
        {
            if (m_bits == 0x1) {
                set_ticks(m_computer_char_ticks);
                m_state = KEYER_STATE_CHAR_SPACE;
                protocol.kr_computer_echo();
                break;
            }
                  
            if (m_bits & 1) {
                set_ticks(m_computer_dah_down_ticks);       
                m_state = KEYER_STATE_DOWN_DIT;
            }
            else {
                set_ticks(m_computer_dit_down_ticks);
                m_state = KEYER_STATE_DOWN_DAH;
            }

            KEY_DOWN;
            m_bits >>= 1;
            break;
        }

        // Completed a word or special space
        case KEYER_STATE_SPACE:
        {
            set_ticks(m_computer_char_ticks);
            m_state = KEYER_STATE_CHAR_SPACE;
            protocol.kr_computer_echo();
            break;
        }

        // Completed character space
        case KEYER_STATE_CHAR_SPACE:
            send_or_idle_computer();
            break;

        case KEYER_STATE_TUNE:
        {
            KEY_DOWN;
            set_deadman();
            protocol.kr_state(PC_STATE_TUNING);
            break;
        }
  
        default:
            error.bad_state("computer_timer", m_state, m_next_state);
    }
}

////////////////// Manual Keyer Interface Routines /////////////////////

 // These are for the for the keyer_manual subclasses
void
keyer_keyer::km_send_dit()
//----------------------------------------------------------------------
// Send a DIT from the paddle
//----------------------------------------------------------------------
{
    start_paddle(KEYER_STATE_DOWN_DIT);
}

void
keyer_keyer::km_send_dah()
//----------------------------------------------------------------------
// Send a DAH from the paddle
//----------------------------------------------------------------------
{
    start_paddle(KEYER_STATE_DOWN_DAH);
}

void
keyer_keyer::km_send_space()
//----------------------------------------------------------------------
// Send an inter-character space from the paddle (autospace)
//----------------------------------------------------------------------
{
    set_ticks(3 * m_base_ticks);
    m_state = KEYER_STATE_CHAR_SPACE;
}

void
keyer_keyer::km_start_sending()
//----------------------------------------------------------------------
// Set the key down from the paddle (bug DAH or straight key)
//----------------------------------------------------------------------
{
    start_paddle(KEYER_STATE_DOWN);
}

void
keyer_keyer::km_stop_sending()
//----------------------------------------------------------------------
// Stop sending from the paddle
//----------------------------------------------------------------------
{
    if (!m_paddle) {
        error.general("keyer_keyer::km_stop_sending called when not in paddle mode");
        return;
    }

    if ((m_state == KEYER_STATE_DOWN) || (m_state == KEYER_STATE_DOWN_DIT)
    || (m_state == KEYER_STATE_DOWN_DAH)) {
        KEY_UP;
    }

    // If the state is key down there is a deadman timer running
    // for it.  That timer must be canceled because it has a long
    // timeout and could fill the queue.
    if (m_state == KEYER_STATE_DOWN) {
        queue.cancel_timed(m_deadman_timer);
    }

    idle_paddle_keyer();
}

void
keyer_keyer::km_alter_speed(int32_t delta)
//----------------------------------------------------------------------
// Change the speed while sending from the paddle
//----------------------------------------------------------------------
{
    m_alter_speed = delta;
    update_speed();
}

void
keyer_keyer::km_alter_weight(int32_t delta)
//----------------------------------------------------------------------
// Change the weight while sending from the paddle
//----------------------------------------------------------------------
{
    m_alter_weight = delta;
    update_speed();
}

void
keyer_keyer::km_idle_sync()
//----------------------------------------------------------------------
// Synchronous wait (for real Ultimatic)
//----------------------------------------------------------------------
{
    set_ticks(m_base_ticks);
    m_state = KEYER_STATE_IDLE_SYNC;

    // If the keyer doesn't send for a while a space character should
    // be echoed.
    if (m_sync_idles < 2) {
        m_sync_idles++;
        if (m_sync_idles == 2) {
            if (m_echo) {
                protocol.kr_echo(' ');
            }
        }
    }
}

////////////////// PC Interface Routines ///////////////////////////////
void
keyer_keyer::pc_keyer_type(keyer_type type)
//----------------------------------------------------------------------
// PC sets the keyer type
//----------------------------------------------------------------------
{
    if (config.get_keyer_local()) {
        return;
    }

    set_keyer_type(type, true);
}

void
keyer_keyer::pc_paddle_swap(bool reverse)
//----------------------------------------------------------------------
// PC reverses the dit and dah paddles
//----------------------------------------------------------------------
{
    if (config.get_keyer_local()) {
        return;
    }

    if (reverse) {
        config.set_keyer_paddle_swap(1);
    }
    else {
        config.set_keyer_paddle_swap(1);
    }
}

void
keyer_keyer::pc_autospace(bool autospace)
//----------------------------------------------------------------------
// PC turns autospace on or off
//----------------------------------------------------------------------
{
    if (config.get_keyer_local()) {
        return;
    }

    config.set_keyer_autospace(autospace);
}

void
keyer_keyer::pc_space_6(bool space_6)
//----------------------------------------------------------------------
// PC sets "CT" word spacing (6 dits) or notmal (7 dits)
//----------------------------------------------------------------------
{
    config.set_keyer_space_6(space_6);
}

void
keyer_keyer::pc_weight(uint32_t weight)
//----------------------------------------------------------------------
// PC sets the keyer weight
//----------------------------------------------------------------------
{
    config.set_keyer_weight(weight);
    update_speed();
}

void
keyer_keyer::pc_ptt_pre(uint32_t ptt_pre)
//----------------------------------------------------------------------
// PC sets the PTT pre-send time
//----------------------------------------------------------------------
{
    config.set_keyer_ptt_pre(ptt_pre);
}

void
keyer_keyer::pc_ptt_post(uint32_t ptt_post)
//----------------------------------------------------------------------
// PC sets the PTT post-send time
//----------------------------------------------------------------------
{
    config.set_keyer_ptt_post(ptt_post);
}

void
keyer_keyer::pc_sets_speed(bool sets_speed)
//----------------------------------------------------------------------
// PC sets whether or not the PC controls the keyer speed
//----------------------------------------------------------------------
{
    if (sets_speed) {
        m_computer_sets_speed = true;
        if (!m_computer_speed) {
            m_computer_speed = config.get_keyer_encoder();
        }
    }
    else {
        m_computer_sets_speed = false;
    }
    update_speed();
}

void
keyer_keyer::pc_speed(uint32_t speed)
//----------------------------------------------------------------------
// PC sets the keyer speed
//----------------------------------------------------------------------
{
    m_computer_speed = speed;
    update_speed();
}

void
keyer_keyer::pc_encoder_limits(uint32_t min, uint32_t max)
//----------------------------------------------------------------------
// PC sets the pot (encoder) min and max speeds
//----------------------------------------------------------------------
{
    if (min < KEYER_MIN_SPEED) {
        min = KEYER_MIN_SPEED;
    }
    config.set_keyer_encoder_min(min);

    if (max > KEYER_MAX_SPEED) {
        max = KEYER_MAX_SPEED;
    }
    config.set_keyer_encoder_max(max);
}

void
keyer_keyer::pc_farnsworth(uint32_t farnsworth)
//----------------------------------------------------------------------
// PC sets the Farnsworth speed
//----------------------------------------------------------------------
{
    // If the speed is out of range turn off Farnsworth
    if ((farnsworth < KEYER_MIN_SPEED) ||
        (farnsworth > KEYER_MAX_SPEED)) {
        farnsworth = 0;
    }

    config.set_keyer_farnsworth(farnsworth);
    update_speed();
}

void
keyer_keyer::pc_abort()
//----------------------------------------------------------------------
// PC aborts sending
//----------------------------------------------------------------------
{
    stop_computer();
}

void
keyer_keyer::pc_tune(bool tune)
//----------------------------------------------------------------------
// PC closes or opens the keyer output
//----------------------------------------------------------------------
{
    if (tune) {
        if (m_paddle && ((m_state != KEYER_STATE_IDLE) &&
            (m_state != KEYER_STATE_IDLE_SYNC))) {
            protocol.kr_state(PC_STATE_BUSY);
        }
        else {
            start_computer(KEYER_STATE_TUNE);
        }
    }
    else {
        if (m_state == KEYER_STATE_TUNE) {
            stop_computer();
        }
    }
}

void
keyer_keyer::pc_send(uint8_t character)
//----------------------------------------------------------------------
// PC sends a character
//----------------------------------------------------------------------
{
    if (m_paddle && ((m_state != KEYER_STATE_IDLE) &&
        (m_state != KEYER_STATE_IDLE_SYNC))) {
        protocol.kr_state(PC_STATE_BUSY);
        return;
    }

    m_char = character;

    // Protocol requires that the state change before this routine
    // returns.  Otherwise it will believe that the keyer has gone
    // idle.
    protocol.kr_state(PC_STATE_SENDING);

    if (m_paddle) {
        start_computer(KEYER_STATE_IDLE);
    }
}

void
keyer_keyer::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle a keyer message
//----------------------------------------------------------------------
{
    switch (MESSAGE_ID(message)) {  
        case MSG_CAP_LEFT_CHANGED:
        {
            m_manual->paddle_change(MESSAGE_ID(message), data);
            break;
        }

        case MSG_CAP_RIGHT_CHANGED:
        {
            m_manual->paddle_change(MESSAGE_ID(message), data);
            break;
        }

        case MSG_PHY_LEFT_CHANGED:
        {
            m_manual->paddle_change(MESSAGE_ID(message), data);
            break;
        }

        case MSG_PHY_RIGHT_CHANGED:
        {
            m_manual->paddle_change(MESSAGE_ID(message), data);
            break;
        }

        case MSG_TIMER:
        {
            if (data != m_sequence) {
                return;
            } 

            if (m_paddle) {
                paddle_timer();
            }
            else {
                computer_timer();
            }
            break;
        }

        case MSG_PTT_TIMER:
        {
            if (data != m_ptt_sequence) {
                return;
            }

            if (m_paddle) {
                if ((m_state == KEYER_STATE_IDLE) ||
                    (m_state == KEYER_STATE_IDLE_SYNC)) {
                    PTT_OFF;
                }
            }
            break;
        }

        case MSG_DEADMAN_TIMER:
        {
            if (data != m_deadman_sequence) {
                return;
            }

            if (m_paddle && (m_state == KEYER_STATE_DOWN)) {
                km_stop_sending();
            }

            if (m_state == KEYER_STATE_TUNE) {
                stop_computer();
            }
            break;
        }

        case MSG_SPEED_UP:
        {
            uint8_t speed = config.get_keyer_encoder();
            speed++;
            if (speed <= config.get_keyer_encoder_max()) {
                config.set_keyer_encoder(speed);
                protocol.kr_pot_speed(speed);
                update_speed();
            }
            break;
        }

        case MSG_SPEED_DOWN:
        {
            uint8_t speed = config.get_keyer_encoder();
            speed--;
            if (speed >= config.get_keyer_encoder_min()) {
                config.set_keyer_encoder(speed);
                protocol.kr_pot_speed(speed);
                update_speed();
            }
            break;
        }
        
        case MSG_TYPE_UP:
        {
            keyer_type type = (keyer_type)(config.get_keyer_type() + 1);
            if (type >= KEYER_TYPE_FENCE) {
                type = KEYER_TYPE_FIRST;
            }
            set_keyer_type(type, false);
            break;
        }

        case MSG_TYPE_DOWN:
       {
            keyer_type type = (keyer_type)config.get_keyer_type();
            if (type == KEYER_TYPE_FIRST) {
                type = KEYER_TYPE_FENCE;
            }
            type = (keyer_type)(type - 1);

            set_keyer_type(type, false);
            break;
        }

        case MSG_TYPE:
        {
            set_keyer_type((keyer_type)config.get_keyer_type(), true);
            break;
        }

        case MSG_PADDLE_SWAP:
        {
            if (config.get_keyer_paddle_swap()) {
                config.set_keyer_paddle_swap(0);
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "NOR");
            }
            else {
                config.set_keyer_paddle_swap(1);
                queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "REV");
            }
            break;
        }

        case MSG_RECALC:
        {
            update_speed();
            break;
        }

        default:
            error.bad_message("Keyer", message, data);
    }
}

keyer_keyer keyer;
