/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

// This contains the keyer internal messages

#ifndef KEYER_MESSAGE_H
#define KEYER_MESSAGE_H

#include <wirish.h>

// Messages are 32 bit items containing 8 bits of message class and
// 24 bits of message ID.

#define MESSAGE_ID_WIDTH 24
#define MESSAGE_ID_MASK 0xffffff

#define MESSAGE_CLASS(_message) ((_message) >> MESSAGE_ID_WIDTH)
#define MESSAGE_ID(_message) ((_message) & MESSAGE_ID_MASK)
#define MESSAGE(_class, _id) (uint32_t)((((_class) << MESSAGE_ID_WIDTH) | (_id)))

enum message_class {
    CLASS_UNUSED = 0,     // Should not happen
    CLASS_ENC1,           // Encoder 1
    CLASS_ENC2,           // Encoder 2
    CLASS_OUTPUTS,        // Key and PTT out
    CLASS_BEEP,           // Sidetone
    CLASS_CONTROL,        // Front panel control
    CLASS_MPR121,         // Capacitive paddle
    CLASS_KEYER,          // Big keyer
    CLASS_WK,             // Winkey emulation
    CLASS_STATUS,         // Status indicator keyer
    CLASS_CONFIG,         // Configuration
    CLASS_COMM,           // Serial (USB) communication
    CLASS_PROTOCOL        // Computer protocol
};

enum message_id {
    MSG_UNUSED,           // Should not happen

    MSG_ENCODER_INTERRUPT,// Encoder was rotated
    MSG_BUTTON_DOWN,      // Encoder button pressed
    MSG_BUTTON_UP,        // Encoder button released
    MSG_ENCODER_CCW,      // Encoder rotated CCW
    MSG_ENCODER_CW,       // Encoder rotated CW
    MSG_BUTTON_DEBOUNCE,  // Encoder button debounce timer has expired
    MSG_ENCODER_DEBOUNCE, // Encoder rotation debounce timer has expired

    MSG_CAP_LEFT_CHANGED, // Left capacitive paddle changed, value 0=open, 1=closed
    MSG_CAP_RIGHT_CHANGED,// Right capacitive paddle changed, value 0=open, 1=closed
    MSG_PHY_LEFT_CHANGED, // Left physical paddle changed, value 0=open, 1=closed
    MSG_PHY_RIGHT_CHANGED,// Right physical paddle changed, value 0=open, 1=closed

    MSG_BEEP_OWNER,       // Owner of sidetone, value 0=keyer, 1=status
    MSG_BEEP_ON,          // Turn on sidetone, value=owner
    MSG_BEEP_OFF,         // Turn off sidetone, value=owner
    MSG_BEEP_UPDATE,      // Update amplitude and frequency, value=owner
    MSG_FREQUENCY_UP,     // Increase the frequency, value=owner
    MSG_FREQUENCY_DOWN,   // Decrease the frequency, value=owner
    MSG_VOLUME_UP,        // Increase the volume, value=owner
    MSG_VOLUME_DOWN,      // Decrease the volume, value=owner
    MSG_VALUES,           // Send frequency and volume to status keyer
    MSG_VALUES_XVOL,      // Send status volume to status keyer
    MSG_VALUES_XFREQ,     // Send status frequency to status keyer 
    MSG_RECALC,           // Recalculate due to configuration change

    MSG_PADDLE_SWAP,      // Paddle left/right swap
    MSG_SPEED_UP,         // Keyer speed increase
    MSG_SPEED_DOWN,       // Keyer speed decrease
    MSG_TYPE_UP,          // Keyer type (iambic, bug, etc) up
    MSG_TYPE_DOWN,        // Keyer type (iambic, bug, etc) down
    MSG_TYPE,             // Keyer type configuration changed

    MSG_TIMER,            // Timer expired, value=sequence number
    MSG_PTT_TIMER,        // PTT timer expired, value=sequence number
    MSG_DEADMAN_TIMER,    // Deadman timer expired, value=sequence number

    MSG_STATUS_TEXT,      // Send text, value=pointer to character string
    MSG_STATUS_APPEND,    // Append text, value=pointer to character string
    MSG_STATUS_CLEAR,     // Clear status message
    MSG_STATUS_CHAR,      // Send character, append to existing message, value=character
    MSG_STATUS_NUMBER,    // Send number, append to existing message, value=number
    MSG_STATUS_DIT,       // Send a dit

    MSG_CAP_CHANGED,      // MPR121 Capacitive paddle state changed
    MSG_LONG_PRESS,       // Long button press

    MSG_OUTPUT_CHANGED,   // Keyer output changed, value 0=off, 1=on
    MSG_PTT_CHANGED,      // Keyer PTT changed, value 0=off, 1=on

    MSG_CONFIG_START,     // Configuration modification start
    MSG_PARAM_CHANGE,     // Change parameter, value 0=up, 1=down
    MSG_VALUE_CHANGE,     // Change parameter value, value 0=up, 1=down
    MSG_CONFIG_SET,       // Set the changed value
    MSG_CONFIG_CHANGED,   // A configuration parameter has changed, value=config offset

    MSG_DATA_AVAILABLE,   // Data has been received from the PC
    MSG_SEND_CHAR,        // Send character to PC, value=character
    MSG_SEND_NAME,        // Send string to PC, value=pointer to character string

    MSG_COMM_CHAR,        // A character has arrived, value=character

    MSG_LAST
};

// Constant values for message payload
const uint32_t PADDLE_OPEN = 0;
const uint32_t PADDLE_CLOSED = 1;

const uint32_t OWNER_KEYER = 0;
const uint32_t OWNER_STATUS = 1;

const uint32_t OUTPUT_OFF = 0;
const uint32_t OUTPUT_ON = 1;

const uint32_t CHANGE_NONE = 0;
const uint32_t CHANGE_UP = 1;
const uint32_t CHANGE_DOWN = 2;

#endif
