/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include "keyer_pins.h"
#include "keyer_config.h"
#include "keyer_error.h"
#include "keyer_messages.h"
#include "keyer_mpr121.h"
#include "keyer_queue.h"

// The MPR121 is a capacitive input device.  For the keyer it is set
// to detect quickly and assumes that the touch inputs are metal which
// will see a major capacitive change.
//
// When a paddle is touched or released the chip will interrupt the
// microprocessor which the reads the new values and figures out what
// changed.
//
// Most of the code here is to set up the registers of the MPR121.  Once
// that it done it is easy to interface to.
//
// This class is harware dependent.  It depends on the I/O and interrupt
// subsystems.

static const uint8_t ADDRESS = 0x5A;

static const uint8_t TOUCH_L = 0x00;

// The capacitive inputs
static const uint8_t LEFT_PRESSED = 1;
static const uint8_t RIGHT_PRESSED = 2;

// MPR121 Register Defines
// Based on code from
// http://garagelab.com/profiles/blogs/tutorial-mpr121-capacitive-sensor-with-arduino
static const uint8_t MHD_R = 0x2B;
static const uint8_t NHD_R = 0x2C;
static const uint8_t NCL_R = 0x2D;
static const uint8_t FDL_R = 0x2E;
static const uint8_t MHD_F = 0x2F;
static const uint8_t NHD_F = 0x30;
static const uint8_t NCL_F = 0x31;
static const uint8_t FDL_F = 0x32;
static const uint8_t ELE0_T = 0x41;
static const uint8_t ELE0_R = 0x42;
static const uint8_t ELE1_T = 0x43;
static const uint8_t ELE1_R = 0x44;
static const uint8_t ELE2_T = 0x45;
static const uint8_t ELE2_R = 0x46;
static const uint8_t ELE3_T = 0x47;
static const uint8_t ELE3_R = 0x48;
static const uint8_t ELE4_T = 0x49;
static const uint8_t ELE4_R = 0x4A;
static const uint8_t ELE5_T = 0x4B;
static const uint8_t ELE5_R = 0x4C;
static const uint8_t ELE6_T = 0x4D;
static const uint8_t ELE6_R = 0x4E;
static const uint8_t ELE7_T = 0x4F;
static const uint8_t ELE7_R = 0x50;
static const uint8_t ELE8_T = 0x51;
static const uint8_t ELE8_R = 0x52;
static const uint8_t ELE9_T = 0x53;
static const uint8_t ELE9_R = 0x54;
static const uint8_t ELE10_T = 0x55;
static const uint8_t ELE10_R = 0x56;
static const uint8_t ELE11_T = 0x57;
static const uint8_t ELE11_R = 0x58;
static const uint8_t FIL_CFG = 0x5D;
static const uint8_t ELE_CFG = 0x5E;
static const uint8_t GPIO_CTRL0 = 0x73;
static const uint8_t GPIO_CTRL1 = 0x74;
static const uint8_t GPIO_DATA = 0x75;
static const uint8_t GPIO_DIR = 0x76;
static const uint8_t GPIO_EN = 0x77;
static const uint8_t GPIO_SET = 0x78;
static const uint8_t GPIO_CLEAR = 0x79;
static const uint8_t GPIO_TOGGLE = 0x7A;
static const uint8_t ATO_CFG0 = 0x7B;
static const uint8_t ATO_CFGU = 0x7D;
static const uint8_t ATO_CFGL = 0x7E;
static const uint8_t ATO_CFGT = 0x7F;

keyer_mpr121::keyer_mpr121()
    : m_left_pressed(false), m_right_pressed(false)
{
}

void
keyer_mpr121::setup()
//----------------------------------------------------------------------
// Set up the MPR 121 capacitive touch sensor
//----------------------------------------------------------------------
{
    queue.subscribe(CLASS_MPR121, this);

    Wire.begin();

    // Set up the pins
    set_register(ELE_CFG, 0x00); 

    // Section A - Controls filtering when data is > baseline.
    set_register(MHD_R, 0x01);
    set_register(NHD_R, 0x01);
    set_register(NCL_R, 0x00);
    set_register(FDL_R, 0x00);
    
    // Section B - Controls filtering when data is < baseline.
    set_register(MHD_F, 0x01);
    set_register(NHD_F, 0x01);
    set_register(NCL_F, 0xFF);
    set_register(FDL_F, 0x02);
    
    // Section C - Sets touch and release thresholds for each electrode
    uint8_t touch_threshold = config.get_touch_threshold();
    uint8_t release_threshold = config.get_release_threshold();
    
    set_register(ELE0_T, touch_threshold);
    set_register(ELE0_R, release_threshold);
    
    set_register(ELE1_T, touch_threshold);
    set_register(ELE1_R, release_threshold);

    // The ten dollar keyer only uses two capacitive inputs.  It
    // could be expanded.  Additional inputs could be used for
    // memory playback or other control features.
 #ifdef ADDITIONAL_CAPACITIVE_INPUTS
    set_register(ELE2_T, touch_threshold);
    set_register(ELE2_R, release_threshold);
    
    set_register(ELE3_T, touch_threshold);
    set_register(ELE3_R, release_threshold);
    
    set_register(ELE4_T, touch_threshold);
    set_register(ELE4_R, release_threshold);
    
    set_register(ELE5_T, touch_threshold);
    set_register(ELE5_R, release_threshold);
    
    set_register(ELE6_T, touch_threshold);
    set_register(ELE6_R, release_threshold);
    
    set_register(ELE7_T, touch_threshold);
    set_register(ELE7_R, release_threshold);
    
    set_register(ELE8_T, touch_threshold);
    set_register(ELE8_R, release_threshold);
    
    set_register(ELE9_T, touch_threshold);
    set_register(ELE9_R, release_threshold);
    
    set_register(ELE10_T, touch_threshold);
    set_register(ELE10_R, release_threshold);
    
    set_register(ELE11_T, touch_threshold);
    set_register(ELE11_R, release_threshold);
#endif
    
    // Section D
    // Set the Filter Configuration to be as fast as possible
    set_register(FIL_CFG, 0x00);
    
    // Section E
    // Electrode Configuration - Two electrodes are used for the padle
    // Set ELE_CFG to 0x00 to return to standby mode
    set_register(ELE_CFG, 0x02);

    // Section F
    //set_register(ELE_CFG, 0x0C);

    // Set up the intterupt request pin
    pinMode(MPR121_INT, INPUT_PULLUP);
    attachInterrupt(MPR121_INT, keyer_mpr121::interrupt, FALLING);
}

void
keyer_mpr121::interrupt()
//----------------------------------------------------------------------
// Interrupt callback for capacitive sensor changes
//----------------------------------------------------------------------
{
    queue.put(MESSAGE(CLASS_MPR121, MSG_CAP_CHANGED), 0);  
}

void
keyer_mpr121::set_register(uint8_t reg, uint8_t val)
//----------------------------------------------------------------------
// Set an MPR121 register via I2C
//----------------------------------------------------------------------
{
    Wire.beginTransmission(ADDRESS);
    Wire.write(reg);
    Wire.write(val);
    Wire.endTransmission();
}

uint8_t
keyer_mpr121::get_register(uint8_t reg)
//----------------------------------------------------------------------
// Get an MPR1212 register via I2C
//----------------------------------------------------------------------
{
    Wire.beginTransmission(ADDRESS);
    Wire.write(reg);
    Wire.endTransmission();
    while (Wire.requestFrom(ADDRESS, 1) != 1);
    uint8_t data = Wire.read();
    return data;
}

void
keyer_mpr121::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle capacitive sensor change messages
//----------------------------------------------------------------------
{
        switch (MESSAGE_ID(message)) {
        case MSG_CAP_CHANGED:
        {
            uint8_t data =  get_register(TOUCH_L);
            if (data & LEFT_PRESSED) {
                if (!m_left_pressed) {
                    m_left_pressed = true;
                    queue.put(MESSAGE(CLASS_KEYER, MSG_CAP_LEFT_CHANGED), PADDLE_CLOSED);
                }
            }
            else {
                if (m_left_pressed) {
                    m_left_pressed = false;
                    queue.put(MESSAGE(CLASS_KEYER, MSG_CAP_LEFT_CHANGED), PADDLE_OPEN);
                }
            }
            if (data & RIGHT_PRESSED) {
                if (!m_right_pressed) {
                    m_right_pressed = true;
                    queue.put(MESSAGE(CLASS_KEYER, MSG_CAP_RIGHT_CHANGED), PADDLE_CLOSED);
                }
            }
            else {
                if (m_right_pressed) {
                    m_right_pressed = false;
                    queue.put(MESSAGE(CLASS_KEYER, MSG_CAP_RIGHT_CHANGED), PADDLE_OPEN);
                }
            }
            break;
        }

        default:
            error.bad_message("MPR121", message, data);
    }
}
keyer_mpr121 mpr121;
