/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

// Keyer protocol emulation

// If you use this code in another project or you use it as the basis
// of a new protocol emulator please "Reference the Ten Dollar Keyer"
// in any publicly available code or documentation.

// This file contains code which handles keyer commands  from a serial
// connection.  The commands are compatible  with the commands for the
// K1EL Winkey keyer version 1.
//
// In addition to the normal modality of a keyer, the  emulated keyer has
// am "open" and "closed" mode.  Since it is a serial protocol there is no
// way to tell what the state of the connection is.  For example the
// logging program could restart or the power to the keyer could be
// interrupted.
//
// The keyer mode will be set to "open" if a command is received which can
// only be handled in "open" state.
//
// Some of the commands are multiple bytes.  This emulator does assume that
// it will get an entire command, that is, no partial command will be in the
// buffer when it connects.
//
// The Winkey protocol has limitations.  There is no way to read many values,
// such as the keyer type or the weight.  And the commands which set things
// such as echo also set other thing such as the keyer type, which makes it
// impossible for the PC to keep the values set on the keyer.  Also there is
// no way to tell whether the device is a Winkey or an emulation, and if an
// emulation, which one.
//
// Two extensions to the Winkey protocol are implemented here.  The first
// provides a way to tell that this is a ten dollar keyer.  The second adds
// commands to get and set configuration parameters and to be notified if
// a configuration parameter changes.
//
// There is a character echo command in the Winkey.  If the command is used
// to echo the characters "10$K" this keyer will reply with "10$k" (lower
// case K) while a Winkey or other emulation will respond with an upper case
// K.  This can be used to detect this keyer.
//
// The extensions use commands that are not used by the Winkey.  These allow
// query, set, and test of all configuration parameters.  Each parameter has
// a four character name.
//
// The commands are 
//
//   0x70 ('p') Test for a configuration entry
//   The command is followed by four characters which are the parameter
//   name.  The keyer responds with 0x70, the parameter name, and 1 if it
//   exists or zero if it does not.
//
//   For example if the PC sent 0x70 'W' 'X' 'Y' 'Z' the keyer would
//   respond 0x70 'W' 'X' 'Y' 'Z' 0x00 because there is no WXYZ configuration
//   parameter.
//
//   0x71 ('q') Get a configuration value
//   The command is followed by four characters which are the parameter
//   name.  The keyer responds with 0x71, the parameter name, and the
//   current value of the parameter.
//
//   for example if the PC sent 0x71 'K' 'T' 'Y' 'P' the keyer would
//   respond 0x71 'K' 'T' 'Y' 'P' 0x01 if the current keyer type were
//   Iambic B.
//
//   0x72 ('r') Set a configuration value
//   The command is followed by five characters which are the parameter
//   name and one character which is the new value.  Values may not be
//   range checked (the PC should not send a bad value).  The keyer does
//   not respond.  Unrecognized parameters are ignored.
//
//   For example if the PC sent 0x72 'F' 'R' 'E' 'Q' 0x30 the sidetone
//   frequency would be set to 480 Hz (30 hex is 48 decimal).
//
//   Setting a configuration parameter may have side effects.  For example
//   setting the FREQ parameter changes the sidetone frequency.  There
//   may be some parameters which have side effects that will not occur
//   when set from the PC.
//
//   0x73 ('s') Set status change
//   The command is followed by one character, which if non-zero will
//   turn on status change notification.  If this is on the keyer will
//   send an unsolicited message any time a configuration parameter is
//   changed.  The message will look exactly the same as the message
//   sent in response to a 0x71 configuration value query.  There is no
//   way to select which configuration messages will be sent - the PC
//   can simply ignore those it is not interested in.
//
// The emulation does not implement the full Winkey version 1 protocol.
// It handles commands which are used by logging programs.  If additional
// commands are required they could be added.  This may require adding
// new capabilities to the keyer.
//
// This class is hardware independent.  It also does not know much about
// the rest of the keyer.  This is an attempt to keep the protocol code
// as simple as possible.

#include "keyer_config.h"
#include "keyer_error.h"
#include "keyer_keyer.h"
#include "keyer_messages.h"
#include "keyer_queue.h"
#include "keyer_protocol.h"
#include "keyer_types.h"

// Emulate a version 1 keyer device
static const uint8_t KE_VERSION = 10;

// Emulated keyer status messages
static const uint8_t PC_STATUS_IDLE    = 0xc0;
static const uint8_t PC_STATUS_FULL    = 0xc1;
static const uint8_t PC_STATUS_PADDLE  = 0xc2;
static const uint8_t PC_STATUS_SENDING = 0xc4;
static const uint8_t PC_STATUS_TUNING  = 0xc8;

static const uint32_t DEFAULT_MIN_SPEED = 5;
static const uint32_t DEFAULT_MAX_SPEED = 35;

static const char *MAGIC = "10$K";
static const uint32_t MAGIC_LEN = 4;
static const uint8_t MAGIC_LAST = 'k';

#define SEND(_x) queue.put(MESSAGE(CLASS_COMM, MSG_SEND_CHAR), _x)

keyer_protocol::keyer_protocol()
{
  
}

void 
keyer_protocol::setup()
//----------------------------------------------------------------------
// Set up the protocol instance
//----------------------------------------------------------------------
{
    reset();
    queue.subscribe(CLASS_PROTOCOL, this);
}

void 
keyer_protocol::reset()
//----------------------------------------------------------------------
// Reset to initial state
//----------------------------------------------------------------------
{
    m_state = PC_STATE_IDLE;
    m_insptr = 0;
    m_sendptr = 0;
    m_host_open = false;
    m_keyer_mode = 0;
    m_ultimatic_mode = 0;
    m_echo = false;
    m_overwrite = false;
    m_buflen = 0;
    m_speed = 0;
    m_base_speed = 0;
    m_echo_char = '\0';
    m_sent_status = PC_STATUS_IDLE;
    m_force_status = false;
    m_mode_seen = false;
    m_ultimatic_seen = false;
    m_magic = 0;
    m_extensions = false;
    m_send_config = false;
}

void
keyer_protocol::flush_buffers()
//----------------------------------------------------------------------
// Clear the character and command buffers
//----------------------------------------------------------------------
{
    m_insptr = 0;
    m_sendptr = 0;
    m_overwrite = false;
    m_buflen = 0;
    m_echo_char = '\0';
}

void
keyer_protocol::send_char()
//----------------------------------------------------------------------
// Send the next character, if any
//----------------------------------------------------------------------
{
    uint32_t initial_sendptr = m_sendptr;

    while (m_sendptr < m_buflen) {
        // Ignore buffered NOP and editing remains
        if ((m_buf[m_sendptr] == 0x00) || (m_buf[m_sendptr] == 0x1F)) {
            m_sendptr++;
            continue;
        }

        if (m_buf[m_sendptr] == 0x1C) {
            m_sendptr++;
            m_speed = m_buf[m_sendptr++];
            if (!m_speed) {
                m_speed = m_base_speed;
            }
            keyer.pc_speed(m_speed);
            continue;
        }
        
        if (m_buf[m_sendptr] == 0x1e) {
            m_sendptr++;
            // Set delta to zero
            m_speed = m_base_speed;
            keyer.pc_speed(m_speed);
            continue;
        }

         break; // Real character to send
    }

    if (m_sendptr == m_buflen) {
        // Special case for buffer only containing nothing to send.
        // These get wiped out when they are sent.
        if ((m_state == PC_STATE_IDLE) && (initial_sendptr == 0)) {
            flush_buffers();
        }

        return;
    }

    m_echo_char = m_buf[m_sendptr++];
    keyer.pc_send(m_echo_char);
}

void
keyer_protocol::start_sending()
//----------------------------------------------------------------------
// Start sending characters if appropriate
//----------------------------------------------------------------------
{
    if (m_state != PC_STATE_IDLE) {
        return;
    }

    // Don't start sending if the protocol is in overwrite mode.  N1MM
    // Logger takes advantage of this.
    if (m_overwrite) {
        return;
    }

    send_char();
}

void
keyer_protocol::insert(uint8_t c)
//----------------------------------------------------------------------
// Insert a byte into the buffer
//----------------------------------------------------------------------
{
    // Don't insert characters if the keyer is sending from the
    // paddle
    if (m_state == PC_STATE_BUSY) {
        return;
    }

    if (m_insptr < (BUFFER_SIZE)) {
        m_buf[m_insptr++] = c;
        if (m_insptr > m_buflen) {
            m_buflen = m_insptr;
            send_pc_status();  // In case buffer is now 3/4 full
        }

        start_sending();
    }
    else {
        handle_overflow();
    }
}

void
keyer_protocol::insert(uint8_t c1, uint8_t c2)
//----------------------------------------------------------------------
// Insert two bytes into the buffer
//----------------------------------------------------------------------
{
    // Don't insert characters if the keyer is sending from the
    // paddle
    if (m_state == PC_STATE_BUSY) {
        return;
    }

    if (m_insptr < (BUFFER_SIZE-1)) {
        m_buf[m_insptr++] = c1;
        m_buf[m_insptr++] = c2;
        if (m_insptr > m_buflen) {
            m_buflen = m_insptr;
            send_pc_status();  // In case buffer is now 3/4 full
        }

        start_sending();
    }
    else {
        handle_overflow();
    }
}

void
keyer_protocol::handle_overflow()
//----------------------------------------------------------------------
// Handle a buffer overflow
//----------------------------------------------------------------------
{
    // A character won't fit in the buffer.  This should never
    // happen if the controlling program is properly written.
    // If it depends on the buffer full message it may get into
    // trouble because the USB connection is much faster than the
    // 1200 baud serial connection the protocol was designed for.
    // Fake a paddle interrupt - that should keep the controlling
    // program from wedging up.
    pc_state temp_state =  m_state;
    m_state = PC_STATE_BUSY;
    m_force_status = true;
    send_pc_status();
    m_state = temp_state;
    flush_buffers();
    m_cmdlen = 0;
    send_pc_status();
}

void
keyer_protocol::send_pc_status()
//----------------------------------------------------------------------
// Send a status message to the PC
//----------------------------------------------------------------------
{
    if (!m_host_open) {
        return;
    }

    uint32_t temp = 0;
    switch (m_state) {
        case PC_STATE_IDLE:
            temp = PC_STATUS_IDLE;
            break;

        case PC_STATE_BUSY:
            temp = PC_STATUS_PADDLE | PC_STATUS_SENDING;
            break;

        case PC_STATE_SENDING:
            temp = PC_STATUS_SENDING;
            break;

        case PC_STATE_TUNING:
            temp = PC_STATUS_TUNING;
            break;
    }

    if (m_insptr > (BUFFER_SIZE * 2 / 3)) {
        temp |= PC_STATUS_FULL;
    }

    if (m_force_status || (temp != m_sent_status)) {
        SEND(temp);
        m_sent_status = temp;
        m_force_status = false;
    }
}

void
keyer_protocol::set_mode()
//----------------------------------------------------------------------
// Set the keyer mode from the protocol mode and pin config bits
//----------------------------------------------------------------------
{
    if (m_mode_seen) {
        m_echo = (m_keyer_mode & (1<<2));
        m_paddle_echo = (m_keyer_mode & (1<<6));

        if (m_keyer_mode & (1<<3)) {
            keyer.pc_paddle_swap(true);
        }
        else {
            keyer.pc_paddle_swap(false);
        }

        if (m_keyer_mode & (1<<1)) {
            keyer.pc_autospace(true);
        }
        else {
            keyer.pc_autospace(false);
        }

        if (m_keyer_mode & 1) {
            keyer.pc_space_6(true);
        }
        else {
            keyer.pc_space_6(false);
        }

        switch ((m_keyer_mode >> 4) & 0x3) {
            case 0:
                keyer.pc_keyer_type(KEYER_TYPE_IAMBIC_B);
                break;

            case 1:
                keyer.pc_keyer_type(KEYER_TYPE_IAMBIC_A);
                break;

            case 2:
                if (m_ultimatic_seen) {
                    switch (m_ultimatic_mode >> 6) {
                        case 0:
                            keyer.pc_keyer_type(KEYER_TYPE_ULTIMATIC);
                            break;

                        case 1:
                            keyer.pc_keyer_type(KEYER_TYPE_DAH);
                            break;

                        case 2:                        
                            keyer.pc_keyer_type(KEYER_TYPE_DIT);
                            break;

                        case 3:
                            break;
                    }
                }
                else {
                    keyer.pc_keyer_type(KEYER_TYPE_ULTIMATIC);
                }
                break;

            case 3:
                keyer.pc_keyer_type(KEYER_TYPE_BUG);
                break;
        }   
    }
}

void
keyer_protocol::process_char(uint8_t val)
//----------------------------------------------------------------------
// Process a character read on the serial port
//----------------------------------------------------------------------
{
    bool good_cmd = false;
    bool consume = false;
  
    if (m_cmdlen >= MAX_CMD_LEN) {
        handle_overflow();
        return;
    }
  
    m_cmdbuf[m_cmdlen++] = val;
  
    // Parse the commands.  Commands are various lengths
    // and there is an editable buffer.

#define CONTINUE good_cmd = true; break
#define CONSUME consume = true; good_cmd = true; break

    // Non-administrative commands should only be sent when
    // the keyer is open.  If one is received it means data
    // was sent that this program didn't get.  Open the keyer
    if (!m_host_open && m_cmdbuf[0] != 0x00) {
        m_host_open = true;
        keyer.pc_sets_speed(true);
    }
  
    switch (m_cmdbuf[0]) {
  
        case 0x00:  // Admin command
        {
            if (m_cmdlen < 2) {
              CONTINUE;
            }
      
            switch (m_cmdbuf[1]) {

                case 0x00:  // Timing - Ignore it and the 0xFF
                {
                    if (m_cmdlen < 3) {
                        CONTINUE;
                    }
                    CONSUME;
                }
  
                case 0x01: // Reset
                {
                    reset();

                    CONSUME;
                }
  
                case 0x02: // Open
                {  
                    m_host_open = true;
                    SEND(KE_VERSION);
          
                    keyer.pc_sets_speed(true);
  
                    CONSUME;
                }
  
                case 0x03: // Close
                {
                    m_host_open = false;
                    keyer.pc_sets_speed(false);
    
                    CONSUME;
                }
  
                case 0x04: // Echo
                {
                    if (m_cmdlen < 3) {
                      CONTINUE;
                    }

                    // If the PC sends a magic echo sequence then
                    // modify the response so that it will know it
                    // has found this keyer and enable extensions.
                    uint8_t c = m_cmdbuf[2];
                    if (c == MAGIC[m_magic]) {
                        m_magic++;
                        if (m_magic == MAGIC_LEN) {
                            c = MAGIC_LAST;
                            m_extensions = true;
                            m_magic = 0;
                        }
                    }
                    else {
                        m_magic = 0;
                    }
                    SEND(c);

                    CONSUME;
                }
  
                case 0x05: // Paddle A/D
                case 0x06: // Speed A/D
                case 0x07: // Get Values
                case 0x09: // Get Calibration
                {
                    SEND(0);
                    CONSUME;
                }
  
                case 0x08: // Debug
                case 0x13: // 00 13, screwed-up no-op???
                {
                    CONSUME;
                }
            }

            break;
        }
  
        case 0x01: // Sidetone frequency
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }
            CONSUME;
        }

        case 0x02: // Set speed
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            m_base_speed = m_cmdbuf[1];
            m_speed = m_base_speed;
            keyer.pc_speed(m_speed);
            CONSUME;
        }

        case 0x03: // Weight
        {
            if (m_cmdlen < 2) {
              CONTINUE;
            }
  
            keyer.pc_weight(m_cmdbuf[1]);
            CONSUME;
        }

        case 0x04: // Set lead and tail
        {
            if (m_cmdlen < 3) {
                CONTINUE;
            }

            keyer.pc_ptt_pre(m_cmdbuf[1]);
            keyer.pc_ptt_post(m_cmdbuf[2]);
      
            CONSUME;
        }

        case 0x05: // Set up speed pot
        {
            if (m_cmdlen < 4) {
                CONTINUE;
            }
      
            keyer.pc_encoder_limits(m_cmdbuf[1], m_cmdbuf[1] + m_cmdbuf[2]);

            // For now this doesn't change the speed limits because it is
            // an encoder not a potentiometer

            CONSUME;
        }
    
        case 0x06: // Set paused state (not implemented)
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            CONSUME;
        }

        case 0x07: // Get Speed
        {
            // Speed is sent as 0x80 plus six bits of speed offset
            // by the pot minimum.
            SEND(0x80 | (unsigned char)(keyer.pc_get_encoder() - keyer.pc_get_encoder_min()));

            CONSUME;
        }

        case 0x08: // Backspace
        {
            if (m_insptr) {
                m_insptr--;
            }

            CONSUME;
        }

        case 0x09: // Set Pin configuration
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            // The ultimatic modes aren't supposed to be set
            // in WK1 mode but handle them if they appear
            m_ultimatic_mode = m_cmdbuf[1] >> 6;
            m_ultimatic_seen = true;
            set_mode();

            CONSUME;
        }

        case 0x0A: // Clear buffer, tune, pause
        {
            // Logging programs use this to abort sending
            // N1MM Logger sends this before a message even
            // if the keyer is idle, and then sends characters.
            flush_buffers();
            keyer.pc_abort();

            CONSUME;
        }

        case 0x0B: // Set Tune on/off
        {
            if (m_cmdlen < 2) {
              CONTINUE;
            }

            if (m_cmdbuf[1]) {
                keyer.pc_tune(true);
            }
            else {
                keyer.pc_tune(false);
            }

            CONSUME;
        }

        case 0x0C: // Set high speed CW
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            CONSUME;
        }
    
        case 0x0D: // Set Farnsworth
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            keyer.pc_farnsworth(m_cmdbuf[1]);

            CONSUME;
        }

        case 0x0E: // Set Winkey mode
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }
      
            m_keyer_mode = m_cmdbuf[1];
            m_mode_seen = true;
            set_mode();

            CONSUME;
        }
  
        case 0x0F: // Set defaults
        {
            if (m_cmdlen < 16) {
                CONTINUE;
            }
      
            // Byte 1 - Mode register
            m_keyer_mode = m_cmdbuf[1];
            m_mode_seen = true;
      
            // Byte 2 - Speed in WPM
            m_base_speed = m_cmdbuf[2];
            m_speed = m_base_speed;
            keyer.pc_speed(m_speed);
      
            // Byte 3 - Sidetone Frequency
            // Not implemented
      
            // Byte 4 - Weight
            keyer.pc_weight(m_cmdbuf[4]);
      
            // Byte 5 - Lead-in time
            keyer.pc_ptt_pre(m_cmdbuf[5]);

            // Byte 6 - Tail time
            keyer.pc_ptt_post(m_cmdbuf[6]);
      
            // Byte 7 - Min WPM
            // Byte 8 - WPM range
            keyer.pc_encoder_limits(m_cmdbuf[7], m_cmdbuf[8] + m_cmdbuf[7]);

            // Byte 9 - 1st Extension
            // Not implemented
      
            // Byte 10 - Key compensation
            // Not implemented
      
            // Byte 11 - Farnsworth WPM
            keyer.pc_farnsworth(m_cmdbuf[11]);
      
            // Byte 12 - Paddle setpoint
            // Not implemented
      
            // Byte 13 - Dit/Dah ratio
            // Not implemented
      
            // Byte 14 - Pin configuration
            m_ultimatic_mode = m_cmdbuf[14] >> 6;
            m_ultimatic_seen = true;

            // Byte 15 - Pot range
            // Not implemented

            // Several bytes change the mode.  Set it now
            set_mode();

            CONSUME;
        }

        case 0x10: // Set first extension
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            CONSUME;
        }

        case 0x11: // Set key comp
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            CONSUME;
        }

        case 0x12: // Set paddle switchpoint
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            CONSUME;
        }

        case 0x13: // NO-OP
        {
            CONSUME;
        }

        case 0x14: // Software paddle
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            CONSUME;
        }

        case 0x15: // Status
        {
            m_force_status = true;
            send_pc_status();

              CONSUME;
        }
  
        case 0x16: // Buffer pointer mangling
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }
      
            if (m_cmdbuf[1] == 0) { // 0 - Reset input pointer
                // N1MM Logger sends this before each message.
                m_insptr = 0;
                m_buflen = 0;
                memset(m_buf, 0, BUFFER_SIZE);

                CONSUME;
            }
      
            if (m_cmdlen < 3) {
                CONTINUE;
            }
      
            if (m_cmdbuf[1] == 1) { // 1 - Move pointer in overwrite mode
                m_overwrite = true;
                if (m_cmdbuf[2] < BUFFER_SIZE-1) {
                    m_insptr = m_cmdbuf[2];
                    if (m_insptr > 0) {
                        m_insptr--;
                        send_pc_status();
                    }
                }
      
              CONSUME;
            }
  
            if (m_cmdbuf[1] == 2) { // 2 - Move pointer in append mode
                m_overwrite = false;
                if (m_cmdbuf[2] < BUFFER_SIZE) {
                    m_insptr = m_cmdbuf[2];
                    send_pc_status();
                }
                CONSUME;
            }
      
            if (m_cmdbuf[1] == 3) { // 3 - Write NULLs to buffer
                uint32_t i;
        
                for (i=0; i<m_cmdbuf[2]; i++) {
                    if (m_insptr < BUFFER_SIZE) {
                        m_buf[m_insptr++] = '\0';
                    }
                    else {
                        m_insptr = BUFFER_SIZE-1;
                        break;
                    }
                }
        
                if (m_insptr > m_buflen) {
                    m_buflen = m_insptr;
                }

                send_pc_status();

                CONSUME;
            }

            // Unknown subcommand.  Might as well consume it, otherwise
            // the buffer will eventually overflow.
            CONSUME;
        }
  
        case 0x17: // DIT-DAH ratio
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }
  
            CONSUME;
        }

        case 0x18: // Buffered PTT on/off
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }
            // This isn't supported for now.  Following the Winkey
            // description it would only work if keyer PTT is configured
            // off.
            CONSUME;
        }

        case 0x19: // Buffered Key
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            CONSUME;
        }

        case 0x1A: // Buffered pause
        {
            if (m_cmdlen < 2) {
              CONTINUE;
            }
            CONSUME;
        }

        case 0x1B: // Merge letters
        {
            CONSUME;
        }

        case 0x1C: // Buffered speed change
         {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            insert(0x1C, m_cmdbuf[1]);
      
            CONSUME;
        }

        case 0x1D: // High speed speed change
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            CONSUME;
        }

        case 0x1E: // Cancel buffered speed change
        {
            insert(0x1E);
  
            CONSUME;
        }

        case 0x1F: // Buffered no-op
        {
            insert(0x1F);

            CONSUME;
        }


        // WinKey does not use codes 0x61 - 0x7F.  Use some
        // for the ten dollar keyer.
        case 0x70: // Extension - Test for a configuration entry
        {
            if (m_cmdlen < 5) {
                CONTINUE;
            }

            m_cmdbuf[5] = '\0';

            uint32 vname = (m_cmdbuf[1] << 24) + (m_cmdbuf[2] << 16) +  (m_cmdbuf[3] << 8) + m_cmdbuf[4];
            SEND(0x70);
            queue.put(MESSAGE(CLASS_COMM, MSG_SEND_NAME), vname);
            if (config.get_handle(&(m_cmdbuf[1])) == NO_CONFIGURATION) {
                SEND(0);
            }
            else {
                SEND(1);
            }
            CONSUME;
        }

        case 0x71: // Extension - Get a configuration value
        {
            if (m_cmdlen < 5) {
                CONTINUE;
            }

            m_cmdbuf[5] = '\0';

            uint32 vname = (m_cmdbuf[1] << 24) + (m_cmdbuf[2] << 16) +  (m_cmdbuf[3] << 8) + m_cmdbuf[4];
            SEND(0x71);
            queue.put(MESSAGE(CLASS_COMM, MSG_SEND_NAME), vname);
            uint32_t handle = config.get_handle(&(m_cmdbuf[1]));
            if (handle == NO_CONFIGURATION) {
                SEND(0);
            }
            else {
                SEND(config.get_by_handle(handle));
            }
            CONSUME;
        }

        case 0x72: // Extension - Set a configuration value
        {
            if (m_cmdlen < 6) {
                CONTINUE;
            }

            uint32_t handle = config.get_handle(&(m_cmdbuf[1]));
            if (handle != NO_CONFIGURATION) {
              config.set_by_handle(handle, m_cmdbuf[5]);
            }
            CONSUME;
        }

        case 0x73: // Extension - set status change
        {
            if (m_cmdlen < 2) {
                CONTINUE;
            }

            if (m_cmdbuf[1]) {
                m_send_config = true;
            }
            else {
                m_send_config = false;
            }
            CONSUME;
        }

        default:
        {
          
            uint8_t c = m_cmdbuf[0];
            if (c > 96) {  // Out of range
                c = '?';
            }

            insert(c);

            CONSUME;
        }
    } // switch (m_cmdbuf[0])
  
    if (!good_cmd) {
      m_cmdlen = 0; // Hope to resync?
    }
  
    if (consume) {
      m_cmdlen = 0;
    }
}

// The keyer interface
void
keyer_protocol::kr_state(pc_state state)
//----------------------------------------------------------------------
// Handle a keyer state change
//----------------------------------------------------------------------
{
    if (!m_host_open) {
        return;
    }
    
    m_state = state;

    switch (state) {
        case PC_STATE_IDLE:
            // Calling send_char() will cause a state change to BUSY if
            // there is a character to send.  If so there will be a
            // recursive call to this routine with the new state.  Check
            // whether the state has changed and if so return immediately.
            send_char();

            if (m_state != PC_STATE_IDLE) {
                return;
            }

            flush_buffers();
            break;

        case PC_STATE_BUSY:
            // WriteLog sometimes misses a single paddle status.  Send two.
            send_pc_status();
            m_force_status = true;
            break;

        case PC_STATE_SENDING:
            break;

        case PC_STATE_TUNING:
            break;
    }

    send_pc_status();
}

void keyer_protocol::kr_pot_speed(uint32_t speed)
//----------------------------------------------------------------------
// Handle a keyer pot speed change
//----------------------------------------------------------------------
{
    if (!m_host_open) {
        return;
    }

    SEND(0x80 | speed);
}

void
keyer_protocol::kr_echo(uint8_t character)
//----------------------------------------------------------------------
// Echo a character sent with the paddle if desired
//----------------------------------------------------------------------
{
    if (!m_host_open) {
        return;
    }

    if (m_paddle_echo) {
        SEND(character);
    }
}

void keyer_protocol::kr_computer_echo()
//----------------------------------------------------------------------
// Echo a character sent from the computer if desired
//----------------------------------------------------------------------
{
    if (!m_host_open) {
        return;
    }

    // If computer echo is turned on send the character to the PC.
    // Win-Test sends the next character when it sees this.  If the
    // echo is too early it will affect type-ahead and callsign
    // editing.  Too late won't give the PC time to send the next
    // character reliably.  The keyer calls this routine after it
    // has sent the character as it starts the space before the
    // next character.
    if (m_echo) {
        SEND(m_echo_char);
        m_echo_char = '\0';
    }
}



void keyer_protocol::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle incoming messages
//----------------------------------------------------------------------
{
    switch (MESSAGE_ID(message)) {
    
        case MSG_COMM_CHAR:
        {
            process_char(data);
            break;
        }

        case MSG_CONFIG_CHANGED:
        {
            if (m_send_config) {
                // Send an unsolicited configuration item name and value
                uint8_t buffer[5];
                config.get_name_by_handle(data, buffer);
                uint32 vname = (buffer[0] << 24) + (buffer[1] << 16) +  (buffer[2] << 8) + buffer[3];
                SEND(0x71);
                queue.put(MESSAGE(CLASS_COMM, MSG_SEND_NAME), vname);
                SEND(config.get_by_handle(data));
            }
            break;
        }

        default:
            error.bad_message("Protocol", message, data);
    }
}

keyer_protocol protocol;
