/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

// This contains the protocol emulation

#ifndef KEYER_PROTOCOL_H
#define KEYER_PROTOCOL_H

#include <wirish.h>

#include "keyer_receiver.h"

const uint32_t MAX_CMD_LEN = 64;
const uint32_t BUFFER_SIZE = 128;

enum pc_state {
    PC_STATE_IDLE,           // Not sending
    PC_STATE_BUSY,           // Sending from the paddle
    PC_STATE_SENDING,        // Sending from the computer
    PC_STATE_TUNING          // Keyer is key-down tuning
};

class keyer_protocol : keyer_receiver
{
public:

    keyer_protocol();
    void setup();

    // The keyer interface
    void kr_state(pc_state state);
    void kr_pot_speed(uint32_t speed);
    void kr_echo(uint8_t character);
    void kr_computer_echo();

    void do_work(uint32_t message, uint32_t data);

private:

    bool        m_host_open;          // Emulated device is in "open" mode
    uint32_t    m_speed;              // Current speed
    uint32_t    m_base_speed;         // Speed from explicit command

    bool        m_echo;               // Keyer echos computer characters
    bool        m_paddle_echo;        // Keyer echos paddle characters
    uint8_t     m_echo_char;          // Character to echo
    uint8_t     m_sent_status;        // Last status sent to the PC
    bool        m_force_status;       // Send the status even if unchanged

// This is the command buffer for commands receieved from the computer
// Different commands have different lenghts
    uint8_t     m_cmdbuf[MAX_CMD_LEN];
    uint32_t    m_cmdlen;             // Number of characters in command buffer

// This is the buffer for characters to be sent and buffered
// commands.  It is circular.  The WinKey 1 buffer is 32 bytes
// so in practice this one doesn't wrap.

    uint8_t     m_buf[BUFFER_SIZE];
    uint32_t    m_insptr;             // Place to insert in character buffer
    uint32_t    m_buflen;             // Number of chars in buffer
    uint32_t    m_sendptr;            // Pointer to next character to send
    bool        m_overwrite;          // In buffer overwrite mode

    uint32_t    m_keyer_mode;         // Keyer mode
    bool        m_mode_seen;          // Keyer mode command has been seen
    uint32_t    m_ultimatic_mode;     // Keyer mode from ultimatic bits
    bool        m_ultimatic_seen;     // Ultimatic mode command has been seen

    pc_state    m_state;              // Keyer state info

    uint32_t    m_magic;              // Counter for magic echo string
    bool        m_extensions;         // PC has enabled command extensions
    bool        m_send_config;        // Send config changes to PC

    void reset();
    void set_mode();
    void flush_buffers();
    void send_char();
    void insert(uint8_t c);
    void insert(uint8_t c1, uint8_t c2);
    void handle_overflow();
    void start_sending();
    void process_char(uint8_t val);
    void send_pc_status();
};

extern keyer_protocol protocol;
#endif
