/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include "atomic.h"

#include "keyer_error.h"
#include "keyer_messages.h"
#include "keyer_misc.h"
#include "keyer_pins.h"
#include "keyer_queue.h"
#include "keyer_timer.h"

// All messages consist of three parts: the ID of the recever (class),
// the ID of the message, and a 32 bit value.  They are queued on a
// fixed-length queue.  The main loop calls the work routine which
// runs through anything on the queue and calls the work routine for
// any messages found.
//
// All object instances which can receive messages are derived from
// the keyer_receiver class and have a do_work() routine which
// receives messages.  These instances register themselves with the
// queue code at startup.
//
// The sender of a message knows what object will receive it.  And
// each message can have only one receiver.  This is not as powerful
// as the message pumps used in windowing operating systems but it
// is simpler and is adequate for this project.
//
// The routines turn of interrupts as necessary so it is safe to queue
// messages from an interrupt rouitine.  All interrupt routines (except
// those internal to the Arduino / Maple libraries) queue a message.
//
// A message can be set to be queued after a specific time.  The code
// which handles this has a fixed number of timer slots and they are
// allocated to routines which need them.  There are enough for all of
// the code which requires a timer so they are given out but never
// reclaimed.  All timed code (except for the keyer sidetone) uses
// this mechanism.  Only one hardware timer is used.  The hardware
// timer is 16 bits but this interface provides a 32 bit timer which
// simplifies other code because special code is not needed for large
// timer values which can occur, for example, at very slow keyer speeds.

// __start_critical() and __stop_critical() turn interrupts off and on
// respectively and are used to protect code which could be called from
// an ISR.  They save and restore the interrupt state.  It is possible
// that noInterrupts() and Interrupts() would work if all ISRs have the
// same priority but I think the I2C has a higher priority than the rest.

// This class is hardware dependent.  It depends on the timer and
// interrupt processor subsystems.

static const int TIMER_PRESCALE = (CLOCK_MHZ * TIMER_US);

keyer_queue::keyer_queue()
    : m_queue_len(0), m_queue_next(0), m_queue_overflow(false),
      m_timer(0), m_timed_queue_len(0)
{
    memset(&m_timed_queue, 0, sizeof(m_timed_queue));

    for (uint32_t i=0; i<MAX_RECEIVERS; i++) {
        m_receivers[i] = this;
    }

    timer_pause(TIMER4);
    timer_set_reload(TIMER4, 0xffff);
    timer_set_count(TIMER4, 20000);
    timer_set_prescaler(TIMER4, TIMER_PRESCALE);
    keyer_timer_oneshot(TIMER4);
    keyer_timer_count_down(TIMER4);

    timer_attach_interrupt(TIMER4, TIMER_DIER_UIE_BIT, keyer_queue::interrupt);

    // It isn't clear why the timer needs to run here but if it
    // doesn't the first timed entry is lost.
    timer_resume(TIMER4);
}

void
keyer_queue::subscribe(uint32_t message_class, keyer_receiver *receiver)
//----------------------------------------------------------------------
// Add a subscriber to receive queued messages
//----------------------------------------------------------------------
{
    m_receivers[message_class] = receiver;
}

int32_t
keyer_queue::get_timer()
//----------------------------------------------------------------------
// Reserve a timer slot for a message source
//----------------------------------------------------------------------
{
    if (m_timer == MAX_TIMED_ELEMENTS) {
        error.general("Out of timer elements!");
        return -1;
    }
    return m_timer++;
}

void
keyer_queue::put(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle a message to the back of the queue
//----------------------------------------------------------------------
{
    __start_critical();

    if (m_queue_len < MAX_ELEMENTS) {
        uint32_t index = (m_queue_next + m_queue_len) & (MAX_ELEMENTS - 1);
        m_queue[index].message = message;
        m_queue[index].data = data;
        m_queue_len++;
    }
    else {
        // For debugging
        m_queue_overflow = true;
    }
    __stop_critical();
}

void
keyer_queue::put_timed(int32_t timer, uint32_t ticks, uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Add a message which will be queued when the timeout expires
//----------------------------------------------------------------------
{
    if (timer == -1) {
        return;
    }

    if (!ticks) {
        return put(message, data);
    }
    
    uint32_t next_ticks = 0;
    __start_critical();
    timer_pause(TIMER4);

    if (m_timed_queue_len) {
        // It is possible that the timer just finished counting, in which
        // case there is an interrupt pending.  This code does what the
        // interrupt handler does so if the interrupt fired the work would
        // be done twice.  Clear the interrupt to make sure this doesn't
        // happen.
        //keyer_timer_clear_interrupt(TIMER4);

        // Adjust the existing timer queue elements
        uint32_t completed_ticks = m_ticks - timer_get_count(TIMER4);
        
        for (uint32_t i=0; i<MAX_TIMED_ELEMENTS; i++) {
            if (m_timed_queue[i].ticks) {
                m_timed_queue[i].ticks -= completed_ticks;
                if (!m_timed_queue[i].ticks) {
                    put(m_timed_queue[i].element.message, m_timed_queue[i].element.data);
                    m_timed_queue_len--;
                }
                else {
                    if (!next_ticks || (m_timed_queue[i].ticks < next_ticks)) {
                        next_ticks = m_timed_queue[i].ticks;
                    }
                }
            }
        }
    }

    m_timed_queue[timer].ticks = ticks;
    m_timed_queue[timer].element.message = message;
    m_timed_queue[timer].element.data = data;
    m_timed_queue_len++;
    if (!next_ticks || (ticks < next_ticks)) {
        next_ticks = ticks;
    }

    // Restart the timer 
    if (next_ticks < 0xffff) {
        timer_set_count(TIMER4, next_ticks);
        queue.m_ticks = next_ticks;
    }
    else {
        timer_set_count(TIMER4, 0xffff);
        queue.m_ticks = 0xffff;
    }
    timer_resume(TIMER4);
    
    __stop_critical();
}

void
keyer_queue::cancel_timed(int32_t timer)
//----------------------------------------------------------------------
// Remove a timed message
//----------------------------------------------------------------------
{
    __start_critical();
    if (m_timed_queue[timer].ticks) {
        m_timed_queue[timer].ticks = 0;
        m_timed_queue_len--;
    }
    __stop_critical();
}


void
keyer_queue::run()
//----------------------------------------------------------------------
// Send any queued messages to the appropriate message receivers
//----------------------------------------------------------------------
{
    if (m_queue_overflow) {
        m_queue_overflow = false;
        error.general("*** Queue Overflow ***");
    }
    
    __start_critical();
    while (m_queue_len) {
        uint32_t message = m_queue[m_queue_next].message;
        uint32_t data = m_queue[m_queue_next].data;
        m_queue_next = (m_queue_next + 1) & (MAX_ELEMENTS - 1);
        m_queue_len--;
        __stop_critical();
        m_receivers[MESSAGE_CLASS(message)]->do_work(message, data); 
        __start_critical();
    }
    __stop_critical();
}

// Default setup routine for any class which is not subscribed to.
// This should not be called but might be if there are programming
// errors.
void
keyer_queue::setup()
//----------------------------------------------------------------------
// Default setup routine - should never be called
//----------------------------------------------------------------------
{
    error.general("Setup called for unregistered receiver");
}

// Default do_work routine for any class which is not subscribed to.
// This should not be called but might be if there are programming
// errors.
void
keyer_queue::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Default message receiver - should never be called
//----------------------------------------------------------------------
{
    error.bad_message("Queue", message, data);
}

// This handles the timed queue.  It is called from an ISR only so
// a critical section is not needed
void
keyer_queue::interrupt()
//----------------------------------------------------------------------
// Timer interrupt callback
//----------------------------------------------------------------------
{
    if (!queue.m_timed_queue_len) {
        return;
    }

    uint32_t next_ticks = 0;

    for (uint32_t i=0; i<MAX_TIMED_ELEMENTS; i++) {
        if (queue.m_timed_queue[i].ticks) {
            queue.m_timed_queue[i].ticks -= queue.m_ticks;
            if (queue.m_timed_queue[i].ticks) {
                if (!next_ticks || (queue.m_timed_queue[i].ticks < next_ticks)) {
                    next_ticks = queue.m_timed_queue[i].ticks;
                }
            }
            else {
                queue.put(queue.m_timed_queue[i].element.message, queue.m_timed_queue[i].element.data);
                queue.m_timed_queue_len--;
                if (!queue.m_timed_queue_len) {
                    return;
                }
            }
        }
    }

    if (next_ticks) {
        if (next_ticks < 0xffff) {
            timer_set_count(TIMER4, next_ticks);
            queue.m_ticks = next_ticks;
        }
        else {
            timer_set_count(TIMER4, 0xffff);
            queue.m_ticks = 0xffff;
        }
        timer_resume(TIMER4);
    }
}

#ifdef DEBUG
void
keyer_queue::dump()
//----------------------------------------------------------------------
// Dump the queue for debugging
//----------------------------------------------------------------------
{
  // Debugging only, not complete
    __start_critical();


    memcpy(m_debug_queue, m_queue, sizeof(m_queue));
    m_debug_queue_len = m_queue_len;
    m_debug_queue_next = m_queue_next;
    m_debug_queue_overflow = m_queue_overflow;

    memcpy(m_debug_timed_queue, m_timed_queue, sizeof(m_timed_queue));
    m_debug_ticks = m_ticks;
    m_debug_timer = m_timer;
    m_debug_timed_queue_len = m_timed_queue_len;

    __stop_critical();

    for (int i=0; i<MAX_ELEMENTS; i++) {
        
    }
    Serial.print("Queue Len: ");
    Serial.print(m_debug_queue_len, DEC);
    Serial.print("  Next: ");
    Serial.print(m_debug_queue_next, DEC);
    Serial.print("  Overflow: ");
    Serial.println((m_debug_queue_overflow) ? "T" : "F");

    
}
#endif

keyer_queue queue;
