/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

// This contains the internal message queue class

#ifndef KEYER_QUEUE_H
#define KEYER_QUEUE_H

#include <wirish.h>
#include "keyer_receiver.h"

// The timer ticks for the timer queue will be 20 uS.
static const int TIMER_US = 20;
static const int TIMER_PER_MS = (1000 / TIMER_US);

class keyer_queue : keyer_receiver
{

    // The queue length must be a power of 2
    static const uint32_t MAX_ELEMENTS = 128;

    // The number of simultaneous timed puts
    static const uint32_t MAX_TIMED_ELEMENTS = 10;

    // The number of modules which will receive messages
    static const uint32_t MAX_RECEIVERS = 16;

public:

    keyer_queue();

    void setup();
    void subscribe(uint32_t message_class, keyer_receiver *receiver);
    int32_t get_timer();
    void put_timed(int32_t timer, uint32_t ticks, uint32_t message, uint32_t data);

    void cancel_timed(int32_t timer);

    void put(uint32_t message, uint32_t data);

    void put(uint32_t message, const char *data) {
        put(message, (uint32_t)data);
    }

    void put(uint32_t message, const int &data) {
        put(message, (uint32_t)data);
    }

    void put(uint32_t message, const uint8_t *data) {
        put(message, (uint32_t)data);
    }

    // Consider the queue almost full if more than 3/4 of the elements are
    // in use.
    bool almost_full() {
        return (m_queue_len > (MAX_ELEMENTS * 3) / 4);
    }

#ifdef NOT_NOW
// Debugging only
    void dump();
#endif

    void run();
    void do_work(uint32_t message, uint32_t data);
    static void interrupt();

private:

    struct queue_element {
        uint32_t message;
        uint32_t data;
    };

    struct timed_queue_element {
        uint32_t ticks;
        queue_element element;
    };

    // The work queue
    queue_element m_queue[MAX_ELEMENTS];
    uint32_t m_queue_len;
    uint32_t m_queue_next;
    bool m_queue_overflow;

    // The timed event queue
    timed_queue_element m_timed_queue[MAX_TIMED_ELEMENTS];
    uint32_t m_ticks;
    int32_t m_timer;
    uint32_t m_timed_queue_len;

    keyer_receiver *m_receivers[MAX_RECEIVERS];

    // Versions for debugging
    queue_element m_debug_queue[MAX_ELEMENTS];
    uint32_t m_debug_queue_len;
    uint32_t m_debug_queue_next;
    bool m_debug_queue_overflow;

    // The timed event queue
    timed_queue_element m_debug_timed_queue[MAX_TIMED_ELEMENTS];
    uint32_t m_debug_ticks;
    int32_t m_debug_timer;
    uint32_t m_debug_timed_queue_len;
};

extern keyer_queue queue;
#endif
