/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include "keyer_common.h"
#include "keyer_config.h"
#include "keyer_error.h"
#include "keyer_messages.h"
#include "keyer_misc.h"
#include "keyer_queue.h"
#include "keyer_status.h"

// The status keyer is used to give information to the operator, such
// as the current speed and volume.  It is a simple keyer which just
// sends information provided in messages.  It has a speed control but
// no weight or farnsworth.  The output is messages to the sidetone
// generator.
//
// This code is hardware independent.

#define KEY_DOWN queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_ON), OWNER_STATUS); m_key_down = true
#define KEY_UP queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_OFF), OWNER_STATUS); m_key_down = false

keyer_status::keyer_status()
{
    m_sequence = 0;
}

void
keyer_status::setup()
//----------------------------------------------------------------------
// Set up the status message keyer
//----------------------------------------------------------------------
{
    m_message[0] = '\0';
    m_char = 0;
    m_last = 0;
    m_sending = false;
    
    m_timer = queue.get_timer();

    update_speed();

    queue.subscribe(CLASS_STATUS, this);
}

void
keyer_status::update_speed()
//----------------------------------------------------------------------
// Change the status message keyer speed
//----------------------------------------------------------------------
{
    uint32_t speed = config.get_status_speed();
    if (speed < KEYER_MIN_SPEED) {
        speed = KEYER_MIN_SPEED;
    }
    if (speed > KEYER_MAX_SPEED) {
        speed = KEYER_MAX_SPEED;
    }
    m_ticks = TIMER_UNIT / speed;
}

void
keyer_status::set_ticks(uint32_t ticks)
//----------------------------------------------------------------------
// Set a message to be sent after a specified interval
//----------------------------------------------------------------------
{
    m_sequence++;
    queue.put_timed(m_timer, ticks * m_ticks,
                    MESSAGE(CLASS_STATUS, MSG_TIMER),
                    m_sequence);
}

void
keyer_status::send()
//----------------------------------------------------------------------
// Send or continue sending a character
//----------------------------------------------------------------------
{
    if (m_key_down) {
        // Finish dit or dah
        KEY_UP;
        set_ticks(DIT_TICKS);
        return;
    }

    if (m_bits >= 0xf0) {
        // It is a space.  For the status keyer it does not
        // matter which kind (it doesn't use half-space or short space).
        next_char();
        set_ticks(WORD_TICKS_LONG);
        return;
    }

    if (m_bits == 0x1) {
        next_char();
        set_ticks(CHAR_TICKS);
        return;
    }

    if (!m_bits) {
        // Done sending
        m_sending = false;
        m_last = 0;
        m_char = 0;
        queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_OWNER), OWNER_KEYER);
        return;
    }

    // Continue sending this character
    KEY_DOWN;
    if (m_bits & 1) {
        set_ticks(DAH_TICKS);
    }
    else {
        set_ticks(DIT_TICKS);
    }
    m_bits >>= 1;
}

void
keyer_status::next_char()
//----------------------------------------------------------------------
// Send the next character in a message
//----------------------------------------------------------------------
{
    if (m_char < m_last) {
        char next = m_message[m_char++];
        m_bits = morse_char[next-32];
    }
    else {
        m_bits = 0;
    }
}

void
keyer_status::do_work(uint32_t message, uint32_t data)
//----------------------------------------------------------------------
// Handle status keyer messages
//----------------------------------------------------------------------
{
    switch (MESSAGE_ID(message)) {  
        case MSG_STATUS_TEXT:
        {
            m_last = 0;
            m_char = 0;

            // If sending put a space between what was sent and the
            // new message
            if (m_sending) {
                m_message[m_last++] = ' ';
            }
            // Fall into MSG_STATUS_APPEND
        }

        case MSG_STATUS_APPEND:
        {
            for (char *d = (char *)data; *d; d++) {
                if (m_last >= MESSAGE_CHARS) {
                    error.general("Status keyer buffer overflow");
                    break;
                }
                m_message[m_last++] = *d;
            }

            // Start sending the status message if not already sending
            if (!m_sending) {
                m_sending = true;
                next_char();
                queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_OWNER), OWNER_STATUS);
                send();
            }

            break;
        }

        case MSG_STATUS_CLEAR:
       {
            m_last = 0;
            m_char = 0;

            // If sending put a space between what was sent and the new message
            if (m_sending) {
                m_message[m_last++] = ' ';
            }

            break;
       }

        case MSG_STATUS_CHAR:
        {
            if (m_last >= (MESSAGE_CHARS - 1)) {
                error.general("Status keyer buffer overflow");
                break;
            }
            m_message[m_last++] = data;

            // Start sending the status message if not already sending
            if (!m_sending) {
                m_sending = true;
                next_char();
                queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_OWNER), OWNER_STATUS);
                send();
            }

            break;
        }

        case MSG_STATUS_NUMBER:
        {
            uint32_t number = data;
            if (data >= 1000) {
                char digit = (number / 1000) + '0';
                if (m_last >= (MESSAGE_CHARS - 1)) {
                    error.general("Status keyer buffer overflow");
                    break;
                }
                m_message[m_last++] = digit;
                number %= 1000;
            }

             if (data >= 100) {
                char digit = (number / 100) + '0';
                if (m_last >= (MESSAGE_CHARS - 1)) {
                    error.general("Status keyer buffer overflow");
                    break;
                }
                m_message[m_last++] = digit;
                number %= 100;
            }

             if (data >= 10) {
                char digit = (number / 10) + '0';
                if (m_last >= (MESSAGE_CHARS - 1)) {
                    error.general("Status keyer buffer overflow");
                    break;
                }
                m_message[m_last++] = digit;
                number %= 10;
            }

            if (m_last >= (MESSAGE_CHARS - 1)) {
                error.general("Status keyer buffer overflow");
                break;
            }
            m_message[m_last++] = number + '0';

            // Start sending the status message if not already sending
            if (!m_sending) {
                m_sending = true;
                next_char();
                queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_OWNER), OWNER_STATUS);
                send();
            }

            break;
        }


        case MSG_STATUS_DIT:
        {
            if (m_key_down) { 
                return;
            }

            m_last = 0;
            m_char = 0;
            m_bits = 0x2;

            // Start sending the status message if not already sending
            if (!m_sending) {
                m_sending = true;
                queue.put(MESSAGE(CLASS_BEEP, MSG_BEEP_OWNER), OWNER_STATUS);
                send();
            }

            break;
        }
        case MSG_SPEED_UP:
        {
            uint8_t speed = config.get_status_speed();
            if (speed < KEYER_MAX_SPEED) {
                config.set_status_speed(speed + 1);
                update_speed();
            }
            break;
        }

        case MSG_SPEED_DOWN:
        {
            uint8_t speed = config.get_status_speed();
            if (speed > KEYER_MIN_SPEED) {
                config.set_status_speed(speed - 1);
                update_speed();
            }
            break;
        }

        case MSG_TIMER:
        {
            if (data != m_sequence) {
                return;
            } 
            send();
            break;
        }

        default:
            error.bad_message("Status", message, data);
    }
}

keyer_status status;
