/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2016, 2017 Paul Young
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *****************************************************************************/

#include <wirish.h>

#include "keyer_error.h"
#include "keyer_keyer.h"
#include "keyer_queue.h"
#include "keyer_ultimatic.h"


// This provides the manual (paddle) control for the keyer in sychronous
// ultimatic mode.  The code in keyer_keyer calls this through the public
// routines inherited from keyer_manual.  And this code calls the keyer
// through the public routines in that code.

// The actual Ultimatic keyers (QST Feb 1953, April and May 1955, Sept and
// Oct 1960, 73 June 1966) have a synchronous free-running clock.  This code
// emulates those keyers.  The 1953 and 1966 keyers are single-paddle
// (although the 1966 article hints that it is dual-lever).
//
// The memory latch timing is incorrect.  The real Ultimatic latches at the
// key-up part of a dit or dah.  It isn't clear what the ramifications of
// this are - I need to find a real 1955 or 1961 Ultimatic to test the corner
// cases.
//
// The Ultmatic keyers did not have a PTT output.  The emulation is imperfect
// when the PTT pre or post is non-zero but it should be imperceptable for
// reasonable values of PTT pre and post.


// This class is hardware independent.

keyer_ultimatic::keyer_ultimatic()
{
    
}

bool
keyer_ultimatic::setup(keyer_type type, bool silent)
//----------------------------------------------------------------------
// If the keyer type requested is one of the supported types set it up
//----------------------------------------------------------------------
{
    if (type != KEYER_TYPE_REAL_ULTIMATIC) {
        return false;
    }

    if (!silent) {
        queue.put(MESSAGE(CLASS_STATUS, MSG_STATUS_TEXT), "ULTS");
    }

    m_cap_left = false;
    m_cap_right = false;
    m_phy_left = false;
    m_phy_right = false;

    m_left_seen = false;
    m_right_seen = false;

    keyer.km_set_echo(true);

    keyer.km_idle_sync();

    return true;
}

void keyer_ultimatic::paddle_change(uint32_t id, uint32_t data)
//----------------------------------------------------------------------
// Handle a paddle opening or closing
//----------------------------------------------------------------------
{
   switch (id) {
        case MSG_CAP_LEFT_CHANGED:
        {
            if (data) {
                m_cap_left = true;
                m_left_pressed_last = true;
                m_left_seen = true;
            }
            else {
                m_cap_left = false;
            }
    
            break;
        }

        case MSG_CAP_RIGHT_CHANGED:
        {
            if (data) {
                m_cap_right = true;
                m_left_pressed_last = false;
                m_right_seen = true;
            }
            else {
                m_cap_right = false;
            }
    
            break;
        }

        case MSG_PHY_LEFT_CHANGED:
        {
            if (data) {
                m_phy_left = true;
                m_left_pressed_last = true;
                m_left_seen = true;
            }
            else {
                m_phy_left = false;
            }
    
            break;
        }
 
        case MSG_PHY_RIGHT_CHANGED:
        {
            if (data) {
                m_phy_right = true;
                m_left_pressed_last = false;
                m_right_seen = true;
            }
            else {
                m_phy_right = false;
            }
    
            break;
        }
    }
}

void
keyer_ultimatic::state_change(keyer_state state)
//----------------------------------------------------------------------
// Handle a keyer state change
//----------------------------------------------------------------------
{
    switch (state) {
        case KEYER_STATE_IDLE:
        {
            if (left_paddle()) {
                m_left_seen = true;
            }
            if (right_paddle()) {
                m_right_seen = true;
            }
            if (m_left_seen) {
                if (m_right_seen) {
                    if (m_left_pressed_last) {
                        send_left();
                    }
                    else {
                        send_right();
                    }
                }
                else {
                    send_left();
                }
            }
            else {
                if (m_right_seen) {
                    send_right();
                }
                else {
                        m_left_seen = false;
                        m_right_seen = false;
                        keyer.km_idle_sync();
                }
            }
            break;
        }

        default:
            break;
    }
}

keyer_ultimatic ultimatic;
